
try:
    import mysql.connector
    from mysql.connector import Error
    from zcrmsdk.src.com.zoho.api.authenticator.store.token_store import TokenStore
    from zcrmsdk.src.com.zoho.api.authenticator.oauth_token import OAuthToken
    from zcrmsdk.src.com.zoho.crm.api.util.constants import Constants
    from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException

except Exception as e:
    import mysql.connector
    from mysql.connector import Error
    from .token_store import TokenStore
    from ..oauth_token import OAuthToken
    from ....crm.api.util.constants import Constants
    from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException


class DBStore(TokenStore):

    """
    This class to store user token details to the MySQL DataBase.
    """

    def __init__(self, host=Constants.MYSQL_HOST, database_name=Constants.MYSQL_DATABASE_NAME,
                 user_name=Constants.MYSQL_USER_NAME, password="", port_number=Constants.MYSQL_PORT_NUMBER,
                 table_name=Constants.MYSQL_TABLE_NAME):

        """
        Creates a DBStore class instance with the specified parameters.

        Parameters:
            host (str) : A string containing the DataBase host name. Default value is localhost
            database_name (str) : A string containing the DataBase name. Default value is zohooauth
            user_name (str) : A string containing the DataBase user name. Default value is root
            password (str) : A string containing the DataBase password. Default value is an empty string
            port_number (str) : A string containing the DataBase port number. Default value is 3306
        """

        self.__host = host
        self.__database_name = database_name
        self.__user_name = user_name
        self.__password = password
        self.__port_number = port_number
        self.__table_name = table_name

    def get_host(self):
        """
        This is a getter method to get __host.

        Returns:
            string: A string representing __host
        """

        return self.__host

    def get_database_name(self):
        """
        This is a getter method to get __database_name.

        Returns:
            string: A string representing __database_name
        """

        return self.__database_name

    def get_user_name(self):
        """
        This is a getter method to get __user_name.

        Returns:
            string: A string representing __user_name
        """

        return self.__user_name

    def get_password(self):
        """
        This is a getter method to get __password.

        Returns:
            string: A string representing __password
        """
        return self.__password

    def get_port_number(self):
        """
        This is a getter method to get __port_number.

        Returns:
            string: A string representing __port_number
        """

        return self.__port_number

    def get_table_name(self):
        """
        This is a getter method to get __table_name.

        Returns:
            string: A string representing __table_name
        """

        return self.__table_name

    def get_token(self, user, token):
        cursor = None
        try:
            connection = mysql.connector.connect(host=self.__host, database=self.__database_name, user=self.__user_name, password=self.__password, port=self.__port_number)
            try:
                if isinstance(token, OAuthToken):
                    cursor = connection.cursor()
                    query = self.construct_dbquery(user.get_email(), token, False)
                    cursor.execute(query)
                    result = cursor.fetchone()

                    if result is not None:
                        oauthtoken = OAuthToken(client_id=result[2], client_secret=result[3], grant_token=result[6],
                                                 refresh_token=result[4], redirect_url=result[8])
                        oauthtoken.set_id(result[0])
                        oauthtoken.set_access_token(result[5])
                        oauthtoken.set_expires_in(str(result[7]))
                        oauthtoken.set_user_mail(result[1])
                        return oauthtoken

            except Error as ex:
                raise ex

            finally:
                cursor.close() if cursor is not None else None
                connection.close() if connection is not None else None

        except Error as ex:
            raise SDKException(code=Constants.TOKEN_STORE, message=Constants.GET_TOKEN_DB_ERROR, cause=ex)

    def save_token(self, user, token):
        cursor = None
        try:
            connection = mysql.connector.connect(host=self.__host, database=self.__database_name, user=self.__user_name, password=self.__password, port=self.__port_number)

            try:
                if isinstance(token, OAuthToken):
                    token.set_user_mail(user.get_email())
                    self.delete_token(token)
                    cursor = connection.cursor()
                    query = "insert into "+self.__table_name+" (id,user_mail,client_id,client_secret,refresh_token,access_token,grant_token,expiry_time,redirect_url) values (%s,%s,%s,%s,%s,%s,%s,%s,%s);"
                    val = (token.get_id(), user.get_email(), token.get_client_id(), token.get_client_secret(), token.get_refresh_token(), token.get_access_token(), token.get_grant_token(), token.get_expires_in(), token.get_redirect_url())
                    cursor.execute(query, val)
                    connection.commit()

            except Error as ex:
                raise ex

            finally:
                cursor.close() if cursor is not None else None
                connection.close() if connection is not None else None

        except Error as ex:
            raise SDKException(code=Constants.TOKEN_STORE, message=Constants.SAVE_TOKEN_DB_ERROR, cause=ex)

    def delete_token(self, token):
        cursor = None
        try:
            connection = mysql.connector.connect(host=self.__host, database=self.__database_name, user=self.__user_name, password=self.__password, port=self.__port_number)

            try:
                if isinstance(token, OAuthToken):
                    cursor = connection.cursor()
                    query = self.construct_dbquery(token.get_user_mail(), token, True)
                    cursor.execute(query)
                    connection.commit()

            except Error as ex:
                raise ex

            finally:
                cursor.close() if cursor is not None else None
                connection.close() if connection is not None else None

        except Error as ex:
            raise SDKException(code=Constants.TOKEN_STORE, message=Constants.DELETE_TOKEN_DB_ERROR, cause=ex)

    def get_tokens(self):
        cursor = None
        try:
            connection = mysql.connector.connect(host=self.__host, database=self.__database_name, user=self.__user_name, password=self.__password, port=self.__port_number)
            tokens = []

            try:
                cursor = connection.cursor()
                query = 'select * from ' + self.__table_name + ";"
                cursor.execute(query)
                results = cursor.fetchall()

                for result in results:

                    token = OAuthToken(client_id=result[2], client_secret=result[3], grant_token=result[6], refresh_token=result[4], redirect_url=result[8])
                    token.set_id(result[0])
                    token.set_expires_in(str(result[7]))
                    token.set_user_mail(result[1])
                    token.set_access_token(result[5])
                    tokens.append(token)

                return tokens

            except Error as ex:
                raise ex

            finally:
                cursor.close() if cursor is not None else None
                connection.close() if connection is not None else None

        except Error as ex:
            raise SDKException(Constants.TOKEN_STORE, Constants.GET_TOKENS_DB_ERROR, None, ex)

    def delete_tokens(self):
        cursor = None
        try:
            connection = mysql.connector.connect(host=self.__host, database=self.__database_name, user=self.__user_name, password=self.__password, port=self.__port_number)

            try:
                cursor = connection.cursor()
                query = 'delete from ' + self.__table_name + ";"
                cursor.execute(query)
                connection.commit()

            except Error as ex:
                raise ex

            finally:
                cursor.close() if cursor is not None else None
                connection.close() if connection is not None else None
        except Error as ex:
            raise SDKException(Constants.TOKEN_STORE, Constants.DELETE_TOKENS_DB_ERROR, Exception=ex)

    def get_token_by_id(self, id, token):
        cursor = None
        try:
            connection = mysql.connector.connect(host=self.__host, database=self.__database_name, user=self.__user_name, password=self.__password, port=self.__port_number)
            try:
                if isinstance(token, OAuthToken):

                    query = "select * from " + self.__table_name + " where id='" + id + "'"
                    cursor = connection.cursor()
                    cursor.execute(query)
                    results = cursor.fetchall()

                    for result in results:
                        if result[0] == id:

                            oauthtoken = OAuthToken(client_id=result[2], client_secret=result[3], grant_token=result[6], refresh_token=result[4], redirect_url=result[8])
                            oauthtoken.set_id(result[0])
                            oauthtoken.set_access_token(result[5])
                            oauthtoken.set_expires_in(str(result[7]))
                            oauthtoken.set_user_mail(result[1])
                            return oauthtoken

            except Error as ex:
                raise ex

            finally:
                cursor.close() if cursor is not None else None
                connection.close() if connection is not None else None

        except Error as ex:
            raise SDKException(code=Constants.TOKEN_STORE, message=Constants.GET_TOKEN_BY_ID_DB_ERROR, cause=ex)

    def construct_dbquery(self, email, token, is_delete):
        if email is None:
            raise SDKException(Constants.USER_MAIL_NULL_ERROR, Constants.USER_MAIL_NULL_ERROR_MESSAGE)

        query = "delete from " if is_delete is True else "select * from "
        query += self.__table_name + " where user_mail ='" + email + "' and client_id='" + token.get_client_id() + "' and "

        if token.get_grant_token() is not None:
            query += "grant_token='" + token.get_grant_token() + "'"

        else:
            query += "refresh_token='" + token.get_refresh_token() + "'"

        return query



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import APIResponse, CommonAPIHandler, Constants
except Exception:
	from ..exception import SDKException
	from ..util import APIResponse, CommonAPIHandler, Constants


class TerritoriesOperations(object):
	def __init__(self):
		"""Creates an instance of TerritoriesOperations"""
		pass

	def get_territories(self):
		"""
		The method to get territories

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/settings/territories'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_GET)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_READ)
		try:
			from zcrmsdk.src.com.zoho.crm.api.territories.response_handler import ResponseHandler
		except Exception:
			from .response_handler import ResponseHandler
		return handler_instance.api_call(ResponseHandler.__module__, 'application/json')

	def get_territory(self, id):
		"""
		The method to get territory

		Parameters:
			id (int) : An int representing the id

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/settings/territories/'
		api_path = api_path + str(id)
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_GET)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_READ)
		try:
			from zcrmsdk.src.com.zoho.crm.api.territories.response_handler import ResponseHandler
		except Exception:
			from .response_handler import ResponseHandler
		return handler_instance.api_call(ResponseHandler.__module__, 'application/json')



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Segment(object):
	def __init__(self):
		"""Creates an instance of Segment"""

		self.__id = None
		self.__sequence_number = None
		self.__display_label = None
		self.__type = None
		self.__column_count = None
		self.__fields = None
		self.__buttons = None
		self.__key_modified = dict()

	def get_id(self):
		"""
		The method to get the id

		Returns:
			int: An int representing the id
		"""

		return self.__id

	def set_id(self, id):
		"""
		The method to set the value to id

		Parameters:
			id (int) : An int representing the id
		"""

		if id is not None and not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		self.__id = id
		self.__key_modified['id'] = 1

	def get_sequence_number(self):
		"""
		The method to get the sequence_number

		Returns:
			int: An int representing the sequence_number
		"""

		return self.__sequence_number

	def set_sequence_number(self, sequence_number):
		"""
		The method to set the value to sequence_number

		Parameters:
			sequence_number (int) : An int representing the sequence_number
		"""

		if sequence_number is not None and not isinstance(sequence_number, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: sequence_number EXPECTED TYPE: int', None, None)
		
		self.__sequence_number = sequence_number
		self.__key_modified['sequence_number'] = 1

	def get_display_label(self):
		"""
		The method to get the display_label

		Returns:
			string: A string representing the display_label
		"""

		return self.__display_label

	def set_display_label(self, display_label):
		"""
		The method to set the value to display_label

		Parameters:
			display_label (string) : A string representing the display_label
		"""

		if display_label is not None and not isinstance(display_label, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: display_label EXPECTED TYPE: str', None, None)
		
		self.__display_label = display_label
		self.__key_modified['display_label'] = 1

	def get_type(self):
		"""
		The method to get the type

		Returns:
			string: A string representing the type
		"""

		return self.__type

	def set_type(self, type):
		"""
		The method to set the value to type

		Parameters:
			type (string) : A string representing the type
		"""

		if type is not None and not isinstance(type, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: type EXPECTED TYPE: str', None, None)
		
		self.__type = type
		self.__key_modified['type'] = 1

	def get_column_count(self):
		"""
		The method to get the column_count

		Returns:
			int: An int representing the column_count
		"""

		return self.__column_count

	def set_column_count(self, column_count):
		"""
		The method to set the value to column_count

		Parameters:
			column_count (int) : An int representing the column_count
		"""

		if column_count is not None and not isinstance(column_count, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: column_count EXPECTED TYPE: int', None, None)
		
		self.__column_count = column_count
		self.__key_modified['column_count'] = 1

	def get_fields(self):
		"""
		The method to get the fields

		Returns:
			list: An instance of list
		"""

		return self.__fields

	def set_fields(self, fields):
		"""
		The method to set the value to fields

		Parameters:
			fields (list) : An instance of list
		"""

		if fields is not None and not isinstance(fields, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: fields EXPECTED TYPE: list', None, None)
		
		self.__fields = fields
		self.__key_modified['fields'] = 1

	def get_buttons(self):
		"""
		The method to get the buttons

		Returns:
			list: An instance of list
		"""

		return self.__buttons

	def set_buttons(self, buttons):
		"""
		The method to set the value to buttons

		Parameters:
			buttons (list) : An instance of list
		"""

		if buttons is not None and not isinstance(buttons, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: buttons EXPECTED TYPE: list', None, None)
		
		self.__buttons = buttons
		self.__key_modified['buttons'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Reminder(object):
	def __init__(self):
		"""Creates an instance of Reminder"""

		self.__period = None
		self.__unit = None
		self.__key_modified = dict()

	def get_period(self):
		"""
		The method to get the period

		Returns:
			string: A string representing the period
		"""

		return self.__period

	def set_period(self, period):
		"""
		The method to set the value to period

		Parameters:
			period (string) : A string representing the period
		"""

		if period is not None and not isinstance(period, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: period EXPECTED TYPE: str', None, None)
		
		self.__period = period
		self.__key_modified['period'] = 1

	def get_unit(self):
		"""
		The method to get the unit

		Returns:
			string: A string representing the unit
		"""

		return self.__unit

	def set_unit(self, unit):
		"""
		The method to set the value to unit

		Parameters:
			unit (string) : A string representing the unit
		"""

		if unit is not None and not isinstance(unit, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: unit EXPECTED TYPE: str', None, None)
		
		self.__unit = unit
		self.__key_modified['unit'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Connection(object):
	def __init__(self):
		"""Creates an instance of Connection"""

		self.__source_button = None
		self.__target_screen = None
		self.__key_modified = dict()

	def get_source_button(self):
		"""
		The method to get the source_button

		Returns:
			Button: An instance of Button
		"""

		return self.__source_button

	def set_source_button(self, source_button):
		"""
		The method to set the value to source_button

		Parameters:
			source_button (Button) : An instance of Button
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.wizards.button import Button
		except Exception:
			from .button import Button

		if source_button is not None and not isinstance(source_button, Button):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: source_button EXPECTED TYPE: Button', None, None)
		
		self.__source_button = source_button
		self.__key_modified['source_button'] = 1

	def get_target_screen(self):
		"""
		The method to get the target_screen

		Returns:
			Screen: An instance of Screen
		"""

		return self.__target_screen

	def set_target_screen(self, target_screen):
		"""
		The method to set the value to target_screen

		Parameters:
			target_screen (Screen) : An instance of Screen
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.wizards.screen import Screen
		except Exception:
			from .screen import Screen

		if target_screen is not None and not isinstance(target_screen, Screen):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: target_screen EXPECTED TYPE: Screen', None, None)
		
		self.__target_screen = target_screen
		self.__key_modified['target_screen'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Node(object):
	def __init__(self):
		"""Creates an instance of Node"""

		self.__pos_y = None
		self.__pos_x = None
		self.__start_node = None
		self.__screen = None
		self.__key_modified = dict()

	def get_pos_y(self):
		"""
		The method to get the pos_y

		Returns:
			int: An int representing the pos_y
		"""

		return self.__pos_y

	def set_pos_y(self, pos_y):
		"""
		The method to set the value to pos_y

		Parameters:
			pos_y (int) : An int representing the pos_y
		"""

		if pos_y is not None and not isinstance(pos_y, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: pos_y EXPECTED TYPE: int', None, None)
		
		self.__pos_y = pos_y
		self.__key_modified['pos_y'] = 1

	def get_pos_x(self):
		"""
		The method to get the pos_x

		Returns:
			int: An int representing the pos_x
		"""

		return self.__pos_x

	def set_pos_x(self, pos_x):
		"""
		The method to set the value to pos_x

		Parameters:
			pos_x (int) : An int representing the pos_x
		"""

		if pos_x is not None and not isinstance(pos_x, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: pos_x EXPECTED TYPE: int', None, None)
		
		self.__pos_x = pos_x
		self.__key_modified['pos_x'] = 1

	def get_start_node(self):
		"""
		The method to get the start_node

		Returns:
			bool: A bool representing the start_node
		"""

		return self.__start_node

	def set_start_node(self, start_node):
		"""
		The method to set the value to start_node

		Parameters:
			start_node (bool) : A bool representing the start_node
		"""

		if start_node is not None and not isinstance(start_node, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: start_node EXPECTED TYPE: bool', None, None)
		
		self.__start_node = start_node
		self.__key_modified['start_node'] = 1

	def get_screen(self):
		"""
		The method to get the screen

		Returns:
			Screen: An instance of Screen
		"""

		return self.__screen

	def set_screen(self, screen):
		"""
		The method to set the value to screen

		Parameters:
			screen (Screen) : An instance of Screen
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.wizards.screen import Screen
		except Exception:
			from .screen import Screen

		if screen is not None and not isinstance(screen, Screen):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: screen EXPECTED TYPE: Screen', None, None)
		
		self.__screen = screen
		self.__key_modified['screen'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class UserAddress(object):
	def __init__(self):
		"""Creates an instance of UserAddress"""

		self.__user_name = None
		self.__type = None
		self.__email = None
		self.__id = None
		self.__default = None
		self.__key_modified = dict()

	def get_user_name(self):
		"""
		The method to get the user_name

		Returns:
			string: A string representing the user_name
		"""

		return self.__user_name

	def set_user_name(self, user_name):
		"""
		The method to set the value to user_name

		Parameters:
			user_name (string) : A string representing the user_name
		"""

		if user_name is not None and not isinstance(user_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: user_name EXPECTED TYPE: str', None, None)
		
		self.__user_name = user_name
		self.__key_modified['user_name'] = 1

	def get_type(self):
		"""
		The method to get the type

		Returns:
			string: A string representing the type
		"""

		return self.__type

	def set_type(self, type):
		"""
		The method to set the value to type

		Parameters:
			type (string) : A string representing the type
		"""

		if type is not None and not isinstance(type, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: type EXPECTED TYPE: str', None, None)
		
		self.__type = type
		self.__key_modified['type'] = 1

	def get_email(self):
		"""
		The method to get the email

		Returns:
			string: A string representing the email
		"""

		return self.__email

	def set_email(self, email):
		"""
		The method to set the value to email

		Parameters:
			email (string) : A string representing the email
		"""

		if email is not None and not isinstance(email, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: email EXPECTED TYPE: str', None, None)
		
		self.__email = email
		self.__key_modified['email'] = 1

	def get_id(self):
		"""
		The method to get the id

		Returns:
			int: An int representing the id
		"""

		return self.__id

	def set_id(self, id):
		"""
		The method to set the value to id

		Parameters:
			id (int) : An int representing the id
		"""

		if id is not None and not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		self.__id = id
		self.__key_modified['id'] = 1

	def get_default(self):
		"""
		The method to get the default

		Returns:
			bool: A bool representing the default
		"""

		return self.__default

	def set_default(self, default):
		"""
		The method to set the value to default

		Parameters:
			default (bool) : A bool representing the default
		"""

		if default is not None and not isinstance(default, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: default EXPECTED TYPE: bool', None, None)
		
		self.__default = default
		self.__key_modified['default'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



from .button import Button
from .container import Container
from .node import Node
from .api_exception import APIException
from .response_handler import ResponseHandler
from .chart_data import ChartData
from .wizards_operations import WizardsOperations, GetWizardbyIDParam
from .segment import Segment
from .screen import Screen
from .response_wrapper import ResponseWrapper
from .wizard import Wizard
from .connection import Connection



from .action_wrapper import ActionWrapper
from .api_exception import APIException
from .action_response import ActionResponse
from .response_handler import ResponseHandler
from .success_response import SuccessResponse
from .file_body_wrapper import FileBodyWrapper
from .response_wrapper import ResponseWrapper
from .action_handler import ActionHandler
from .body_wrapper import BodyWrapper
from .related_records_operations import RelatedRecordsOperations, GetRelatedRecordsParam, GetRelatedRecordsHeader, DelinkRecordsParam, GetRelatedRecordHeader



try:
    from zcrmsdk.src.com.zoho.crm.api.dc.data_center import DataCenter
except Exception as e:
    from .data_center import DataCenter


class EUDataCenter(DataCenter):

    """
    This class represents the properties of Zoho CRM in EU Domain.
    """

    @classmethod
    def PRODUCTION(cls):

        """
        This method represents the Zoho CRM Production environment in EU domain
        :return: An instance of Environments
        """

        return DataCenter.Environment("https://www.zohoapis.eu", cls().get_iam_url(), cls().get_file_upload_url(), "eu_prd")

    @classmethod
    def SANDBOX(cls):

        """
        This method represents the Zoho CRM Sandbox environment in EU domain
        :return: An instance of Environment
        """

        return DataCenter.Environment("https://sandbox.zohoapis.eu", cls().get_iam_url(), cls().get_file_upload_url(), "eu_sdb")

    @classmethod
    def DEVELOPER(cls):

        """
        This method represents the Zoho CRM Developer environment in EU domain
        :return:  An instance of Environment
        """

        return DataCenter.Environment("https://developer.zohoapis.eu", cls().get_iam_url(), cls().get_file_upload_url(), "eu_dev")

    def get_iam_url(self):
        return "https://accounts.zoho.eu/oauth/v2/token"

    def get_file_upload_url(self):
        return "https://content.zohoapis.eu"



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Choice, Constants
except Exception:
	from ..exception import SDKException
	from ..util import Choice, Constants


class RequestWrapper(object):
	def __init__(self):
		"""Creates an instance of RequestWrapper"""

		self.__character_encoding = None
		self.__operation = None
		self.__callback = None
		self.__resource = None
		self.__key_modified = dict()

	def get_character_encoding(self):
		"""
		The method to get the character_encoding

		Returns:
			string: A string representing the character_encoding
		"""

		return self.__character_encoding

	def set_character_encoding(self, character_encoding):
		"""
		The method to set the value to character_encoding

		Parameters:
			character_encoding (string) : A string representing the character_encoding
		"""

		if character_encoding is not None and not isinstance(character_encoding, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: character_encoding EXPECTED TYPE: str', None, None)
		
		self.__character_encoding = character_encoding
		self.__key_modified['character_encoding'] = 1

	def get_operation(self):
		"""
		The method to get the operation

		Returns:
			Choice: An instance of Choice
		"""

		return self.__operation

	def set_operation(self, operation):
		"""
		The method to set the value to operation

		Parameters:
			operation (Choice) : An instance of Choice
		"""

		if operation is not None and not isinstance(operation, Choice):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: operation EXPECTED TYPE: Choice', None, None)
		
		self.__operation = operation
		self.__key_modified['operation'] = 1

	def get_callback(self):
		"""
		The method to get the callback

		Returns:
			CallBack: An instance of CallBack
		"""

		return self.__callback

	def set_callback(self, callback):
		"""
		The method to set the value to callback

		Parameters:
			callback (CallBack) : An instance of CallBack
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.bulk_write.call_back import CallBack
		except Exception:
			from .call_back import CallBack

		if callback is not None and not isinstance(callback, CallBack):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: callback EXPECTED TYPE: CallBack', None, None)
		
		self.__callback = callback
		self.__key_modified['callback'] = 1

	def get_resource(self):
		"""
		The method to get the resource

		Returns:
			list: An instance of list
		"""

		return self.__resource

	def set_resource(self, resource):
		"""
		The method to set the value to resource

		Parameters:
			resource (list) : An instance of list
		"""

		if resource is not None and not isinstance(resource, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: resource EXPECTED TYPE: list', None, None)
		
		self.__resource = resource
		self.__key_modified['resource'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
    from dateutil.tz import tz
    import dateutil.parser
    from zcrmsdk.src.com.zoho.crm.api.util.constants import Constants
    from datetime import date, datetime
except Exception:
    from dateutil.tz import tz
    import dateutil.parser
    from .constants import Constants
    from datetime import date, datetime


class DataTypeConverter(object):

    """
    This class converts JSON value to the expected object type and vice versa.
    """

    pre_converter_map = {}

    post_converter_map = {}

    @staticmethod
    def init():

        """
        The method to initialize the PreConverter and PostConverter lambda functions.
        """

        if len(DataTypeConverter.pre_converter_map) != 0 and len(DataTypeConverter.post_converter_map) != 0:
            return

        DataTypeConverter.add_to_map("String", lambda obj: str(obj), lambda obj: str(obj))
        DataTypeConverter.add_to_map("Integer", lambda obj: int(obj), lambda obj: int(obj))
        DataTypeConverter.add_to_map("Long", lambda obj: int(obj) if str(obj) != Constants.NULL_VALUE else None, lambda obj: int(obj))
        DataTypeConverter.add_to_map("Boolean", lambda obj: bool(obj), lambda obj: bool(obj))
        DataTypeConverter.add_to_map("Float", lambda obj: float(obj), lambda obj: float(obj))
        DataTypeConverter.add_to_map("Double", lambda obj: float(obj), lambda obj: float(obj))
        DataTypeConverter.add_to_map("Date", lambda obj: dateutil.parser.isoparse(obj).date(), lambda obj: obj.isoformat())
        DataTypeConverter.add_to_map("DateTime", lambda obj: dateutil.parser.isoparse(obj).astimezone(tz.tzlocal()), lambda obj: obj.replace(microsecond=0).astimezone(tz.tzlocal()).isoformat())
        DataTypeConverter.add_to_map("Object", lambda obj: DataTypeConverter.pre_convert_object_data(obj), lambda obj: DataTypeConverter.post_convert_object_data(obj))

    @staticmethod
    def pre_convert_object_data(obj):
        return obj

    @staticmethod
    def post_convert_object_data(obj):

        if obj is None:
            return None
        if isinstance(obj, list):
            list_value = []
            for data in obj:
                list_value.append(DataTypeConverter.post_convert_object_data(data))

            return list_value

        elif isinstance(obj, dict):
            dict_value = {}
            for key, value in obj.items():
                dict_value[key] = DataTypeConverter.post_convert_object_data(value)

            return dict_value

        elif isinstance(obj, date):
            return DataTypeConverter.post_convert(obj, Constants.DATE_NAMESPACE)

        elif isinstance(obj, datetime):
            return DataTypeConverter.post_convert(obj, Constants.DATETIME_NAMESPACE)

        else:
            return obj

    @staticmethod
    def add_to_map(name, pre_converter, post_converter):

        """
        This method to add PreConverter and PostConverter instance.
        :param name: A str containing the data type class name.
        :param pre_converter: A pre_converter instance.
        :param post_converter: A post_converter instance.
        """

        DataTypeConverter.pre_converter_map[name] = pre_converter
        DataTypeConverter.post_converter_map[name] = post_converter

    @staticmethod
    def pre_convert(obj, data_type):

        """
        The method to convert JSON value to expected data value.
        :param obj: An object containing the JSON value.
        :param data_type: A str containing the expected method return type.
        :return: An object containing the expected data value.
        """

        DataTypeConverter.init()
        return DataTypeConverter.pre_converter_map[data_type](obj)

    @staticmethod
    def post_convert(obj, data_type):

        """
        The method to convert python data to JSON data value.
        :param obj: A object containing the python data value.
        :param data_type: A str containing the expected method return type.
        :return: An object containing the expected data value.
        """

        DataTypeConverter.init()
        return DataTypeConverter.post_converter_map[data_type](obj)



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Button(object):
	def __init__(self):
		"""Creates an instance of Button"""

		self.__id = None
		self.__sequence_number = None
		self.__display_label = None
		self.__criteria = None
		self.__target_screen = None
		self.__type = None
		self.__color = None
		self.__shape = None
		self.__background_color = None
		self.__key_modified = dict()

	def get_id(self):
		"""
		The method to get the id

		Returns:
			int: An int representing the id
		"""

		return self.__id

	def set_id(self, id):
		"""
		The method to set the value to id

		Parameters:
			id (int) : An int representing the id
		"""

		if id is not None and not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		self.__id = id
		self.__key_modified['id'] = 1

	def get_sequence_number(self):
		"""
		The method to get the sequence_number

		Returns:
			int: An int representing the sequence_number
		"""

		return self.__sequence_number

	def set_sequence_number(self, sequence_number):
		"""
		The method to set the value to sequence_number

		Parameters:
			sequence_number (int) : An int representing the sequence_number
		"""

		if sequence_number is not None and not isinstance(sequence_number, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: sequence_number EXPECTED TYPE: int', None, None)
		
		self.__sequence_number = sequence_number
		self.__key_modified['sequence_number'] = 1

	def get_display_label(self):
		"""
		The method to get the display_label

		Returns:
			string: A string representing the display_label
		"""

		return self.__display_label

	def set_display_label(self, display_label):
		"""
		The method to set the value to display_label

		Parameters:
			display_label (string) : A string representing the display_label
		"""

		if display_label is not None and not isinstance(display_label, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: display_label EXPECTED TYPE: str', None, None)
		
		self.__display_label = display_label
		self.__key_modified['display_label'] = 1

	def get_criteria(self):
		"""
		The method to get the criteria

		Returns:
			Criteria: An instance of Criteria
		"""

		return self.__criteria

	def set_criteria(self, criteria):
		"""
		The method to set the value to criteria

		Parameters:
			criteria (Criteria) : An instance of Criteria
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.customviews import Criteria
		except Exception:
			from ..custom_views import Criteria

		if criteria is not None and not isinstance(criteria, Criteria):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: criteria EXPECTED TYPE: Criteria', None, None)
		
		self.__criteria = criteria
		self.__key_modified['criteria'] = 1

	def get_target_screen(self):
		"""
		The method to get the target_screen

		Returns:
			Screen: An instance of Screen
		"""

		return self.__target_screen

	def set_target_screen(self, target_screen):
		"""
		The method to set the value to target_screen

		Parameters:
			target_screen (Screen) : An instance of Screen
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.wizards.screen import Screen
		except Exception:
			from .screen import Screen

		if target_screen is not None and not isinstance(target_screen, Screen):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: target_screen EXPECTED TYPE: Screen', None, None)
		
		self.__target_screen = target_screen
		self.__key_modified['target_screen'] = 1

	def get_type(self):
		"""
		The method to get the type

		Returns:
			string: A string representing the type
		"""

		return self.__type

	def set_type(self, type):
		"""
		The method to set the value to type

		Parameters:
			type (string) : A string representing the type
		"""

		if type is not None and not isinstance(type, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: type EXPECTED TYPE: str', None, None)
		
		self.__type = type
		self.__key_modified['type'] = 1

	def get_color(self):
		"""
		The method to get the color

		Returns:
			string: A string representing the color
		"""

		return self.__color

	def set_color(self, color):
		"""
		The method to set the value to color

		Parameters:
			color (string) : A string representing the color
		"""

		if color is not None and not isinstance(color, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: color EXPECTED TYPE: str', None, None)
		
		self.__color = color
		self.__key_modified['color'] = 1

	def get_shape(self):
		"""
		The method to get the shape

		Returns:
			string: A string representing the shape
		"""

		return self.__shape

	def set_shape(self, shape):
		"""
		The method to set the value to shape

		Parameters:
			shape (string) : A string representing the shape
		"""

		if shape is not None and not isinstance(shape, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: shape EXPECTED TYPE: str', None, None)
		
		self.__shape = shape
		self.__key_modified['shape'] = 1

	def get_background_color(self):
		"""
		The method to get the background_color

		Returns:
			string: A string representing the background_color
		"""

		return self.__background_color

	def set_background_color(self, background_color):
		"""
		The method to set the value to background_color

		Parameters:
			background_color (string) : A string representing the background_color
		"""

		if background_color is not None and not isinstance(background_color, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: background_color EXPECTED TYPE: str', None, None)
		
		self.__background_color = background_color
		self.__key_modified['background_color'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.parameter_map import ParameterMap
	from zcrmsdk.src.com.zoho.crm.api.util import APIResponse, CommonAPIHandler, Constants
	from zcrmsdk.src.com.zoho.crm.api.param import Param
except Exception:
	from ..exception import SDKException
	from ..parameter_map import ParameterMap
	from ..util import APIResponse, CommonAPIHandler, Constants
	from ..param import Param


class TaxesOperations(object):
	def __init__(self):
		"""Creates an instance of TaxesOperations"""
		pass

	def get_taxes(self):
		"""
		The method to get taxes

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/org/taxes'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_GET)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_READ)
		try:
			from zcrmsdk.src.com.zoho.crm.api.taxes.response_handler import ResponseHandler
		except Exception:
			from .response_handler import ResponseHandler
		return handler_instance.api_call(ResponseHandler.__module__, 'application/json')

	def create_taxes(self, request):
		"""
		The method to create taxes

		Parameters:
			request (BodyWrapper) : An instance of BodyWrapper

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.taxes.body_wrapper import BodyWrapper
		except Exception:
			from .body_wrapper import BodyWrapper

		if request is not None and not isinstance(request, BodyWrapper):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: request EXPECTED TYPE: BodyWrapper', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/org/taxes'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_POST)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_CREATE)
		handler_instance.set_content_type('application/json')
		handler_instance.set_request(request)
		handler_instance.set_mandatory_checker(True)
		try:
			from zcrmsdk.src.com.zoho.crm.api.taxes.action_handler import ActionHandler
		except Exception:
			from .action_handler import ActionHandler
		return handler_instance.api_call(ActionHandler.__module__, 'application/json')

	def update_taxes(self, request):
		"""
		The method to update taxes

		Parameters:
			request (BodyWrapper) : An instance of BodyWrapper

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.taxes.body_wrapper import BodyWrapper
		except Exception:
			from .body_wrapper import BodyWrapper

		if request is not None and not isinstance(request, BodyWrapper):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: request EXPECTED TYPE: BodyWrapper', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/org/taxes'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_PUT)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_UPDATE)
		handler_instance.set_content_type('application/json')
		handler_instance.set_request(request)
		handler_instance.set_mandatory_checker(True)
		try:
			from zcrmsdk.src.com.zoho.crm.api.taxes.action_handler import ActionHandler
		except Exception:
			from .action_handler import ActionHandler
		return handler_instance.api_call(ActionHandler.__module__, 'application/json')

	def delete_taxes(self, param_instance=None):
		"""
		The method to delete taxes

		Parameters:
			param_instance (ParameterMap) : An instance of ParameterMap

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if param_instance is not None and not isinstance(param_instance, ParameterMap):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: param_instance EXPECTED TYPE: ParameterMap', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/org/taxes'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_DELETE)
		handler_instance.set_category_method(Constants.REQUEST_METHOD_DELETE)
		handler_instance.set_param(param_instance)
		try:
			from zcrmsdk.src.com.zoho.crm.api.taxes.action_handler import ActionHandler
		except Exception:
			from .action_handler import ActionHandler
		return handler_instance.api_call(ActionHandler.__module__, 'application/json')

	def get_tax(self, id):
		"""
		The method to get tax

		Parameters:
			id (int) : An int representing the id

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/org/taxes/'
		api_path = api_path + str(id)
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_GET)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_READ)
		try:
			from zcrmsdk.src.com.zoho.crm.api.taxes.response_handler import ResponseHandler
		except Exception:
			from .response_handler import ResponseHandler
		return handler_instance.api_call(ResponseHandler.__module__, 'application/json')

	def delete_tax(self, id):
		"""
		The method to delete tax

		Parameters:
			id (int) : An int representing the id

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/org/taxes/'
		api_path = api_path + str(id)
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_DELETE)
		handler_instance.set_category_method(Constants.REQUEST_METHOD_DELETE)
		try:
			from zcrmsdk.src.com.zoho.crm.api.taxes.action_handler import ActionHandler
		except Exception:
			from .action_handler import ActionHandler
		return handler_instance.api_call(ActionHandler.__module__, 'application/json')


class DeleteTaxesParam(object):
	ids = Param('ids', 'com.zoho.crm.api.Taxes.DeleteTaxesParam')



from .api_exception import APIException
from .response_handler import ResponseHandler
from .inventory_templates_operations import InventoryTemplatesOperations, GetInventoryTemplatesParam, GetInventoryTemplatebyIDParam
from .response_wrapper import ResponseWrapper
from .inventory_template import InventoryTemplate



from .request_wrapper import RequestWrapper
from .resource import Resource
from .bulk_write_operations import BulkWriteOperations, UploadFileHeader
from .result import Result
from .api_exception import APIException
from .action_response import ActionResponse
from .response_handler import ResponseHandler
from .file import File
from .call_back import CallBack
from .success_response import SuccessResponse
from .bulk_write_response import BulkWriteResponse
from .file_body_wrapper import FileBodyWrapper
from .field_mapping import FieldMapping
from .response_wrapper import ResponseWrapper



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.parameter_map import ParameterMap
	from zcrmsdk.src.com.zoho.crm.api.util import APIResponse, CommonAPIHandler, Constants
	from zcrmsdk.src.com.zoho.crm.api.param import Param
except Exception:
	from ..exception import SDKException
	from ..parameter_map import ParameterMap
	from ..util import APIResponse, CommonAPIHandler, Constants
	from ..param import Param


class AssignmentRulesOperations(object):
	def __init__(self):
		"""Creates an instance of AssignmentRulesOperations"""
		pass

	def get_assignment_rules(self):
		"""
		The method to get assignment rules

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/settings/automation/assignment_rules'
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_GET)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_READ)
		try:
			from zcrmsdk.src.com.zoho.crm.api.assignment_rules.response_handler import ResponseHandler
		except Exception:
			from .response_handler import ResponseHandler
		return handler_instance.api_call(ResponseHandler.__module__, 'application/json')

	def get_assignment_rule(self, rule_id, param_instance=None):
		"""
		The method to get assignment rule

		Parameters:
			rule_id (int) : An int representing the rule_id
			param_instance (ParameterMap) : An instance of ParameterMap

		Returns:
			APIResponse: An instance of APIResponse

		Raises:
			SDKException
		"""

		if not isinstance(rule_id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: rule_id EXPECTED TYPE: int', None, None)
		
		if param_instance is not None and not isinstance(param_instance, ParameterMap):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: param_instance EXPECTED TYPE: ParameterMap', None, None)
		
		handler_instance = CommonAPIHandler()
		api_path = ''
		api_path = api_path + '/crm/v2.1/settings/automation/assignment_rules/'
		api_path = api_path + str(rule_id)
		handler_instance.set_api_path(api_path)
		handler_instance.set_http_method(Constants.REQUEST_METHOD_GET)
		handler_instance.set_category_method(Constants.REQUEST_CATEGORY_READ)
		handler_instance.set_param(param_instance)
		try:
			from zcrmsdk.src.com.zoho.crm.api.assignment_rules.response_handler import ResponseHandler
		except Exception:
			from .response_handler import ResponseHandler
		return handler_instance.api_call(ResponseHandler.__module__, 'application/json')


class GetAssignmentRuleParam(object):
	module = Param('module', 'com.zoho.crm.api.AssignmentRules.GetAssignmentRuleParam')



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class ImageUpload(object):
	def __init__(self):
		"""Creates an instance of ImageUpload"""

		self.__description = None
		self.__preview_id = None
		self.__encrypted_id = None
		self.__file_name = None
		self.__state = None
		self.__file_id = None
		self.__size = None
		self.__sequence_number = None
		self.__id = None
		self.__key_modified = dict()

	def get_description(self):
		"""
		The method to get the description

		Returns:
			string: A string representing the description
		"""

		return self.__description

	def set_description(self, description):
		"""
		The method to set the value to description

		Parameters:
			description (string) : A string representing the description
		"""

		if description is not None and not isinstance(description, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: description EXPECTED TYPE: str', None, None)
		
		self.__description = description
		self.__key_modified['Description'] = 1

	def get_preview_id(self):
		"""
		The method to get the preview_id

		Returns:
			string: A string representing the preview_id
		"""

		return self.__preview_id

	def set_preview_id(self, preview_id):
		"""
		The method to set the value to preview_id

		Parameters:
			preview_id (string) : A string representing the preview_id
		"""

		if preview_id is not None and not isinstance(preview_id, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: preview_id EXPECTED TYPE: str', None, None)
		
		self.__preview_id = preview_id
		self.__key_modified['Preview_Id'] = 1

	def get_encrypted_id(self):
		"""
		The method to get the encrypted_id

		Returns:
			string: A string representing the encrypted_id
		"""

		return self.__encrypted_id

	def set_encrypted_id(self, encrypted_id):
		"""
		The method to set the value to encrypted_id

		Parameters:
			encrypted_id (string) : A string representing the encrypted_id
		"""

		if encrypted_id is not None and not isinstance(encrypted_id, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: encrypted_id EXPECTED TYPE: str', None, None)
		
		self.__encrypted_id = encrypted_id
		self.__key_modified['Encrypted_Id'] = 1

	def get_file_name(self):
		"""
		The method to get the file_name

		Returns:
			string: A string representing the file_name
		"""

		return self.__file_name

	def set_file_name(self, file_name):
		"""
		The method to set the value to file_name

		Parameters:
			file_name (string) : A string representing the file_name
		"""

		if file_name is not None and not isinstance(file_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: file_name EXPECTED TYPE: str', None, None)
		
		self.__file_name = file_name
		self.__key_modified['File_Name'] = 1

	def get_state(self):
		"""
		The method to get the state

		Returns:
			string: A string representing the state
		"""

		return self.__state

	def set_state(self, state):
		"""
		The method to set the value to state

		Parameters:
			state (string) : A string representing the state
		"""

		if state is not None and not isinstance(state, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: state EXPECTED TYPE: str', None, None)
		
		self.__state = state
		self.__key_modified['State'] = 1

	def get_file_id(self):
		"""
		The method to get the file_id

		Returns:
			string: A string representing the file_id
		"""

		return self.__file_id

	def set_file_id(self, file_id):
		"""
		The method to set the value to file_id

		Parameters:
			file_id (string) : A string representing the file_id
		"""

		if file_id is not None and not isinstance(file_id, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: file_id EXPECTED TYPE: str', None, None)
		
		self.__file_id = file_id
		self.__key_modified['File_Id'] = 1

	def get_size(self):
		"""
		The method to get the size

		Returns:
			int: An int representing the size
		"""

		return self.__size

	def set_size(self, size):
		"""
		The method to set the value to size

		Parameters:
			size (int) : An int representing the size
		"""

		if size is not None and not isinstance(size, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: size EXPECTED TYPE: int', None, None)
		
		self.__size = size
		self.__key_modified['Size'] = 1

	def get_sequence_number(self):
		"""
		The method to get the sequence_number

		Returns:
			int: An int representing the sequence_number
		"""

		return self.__sequence_number

	def set_sequence_number(self, sequence_number):
		"""
		The method to set the value to sequence_number

		Parameters:
			sequence_number (int) : An int representing the sequence_number
		"""

		if sequence_number is not None and not isinstance(sequence_number, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: sequence_number EXPECTED TYPE: int', None, None)
		
		self.__sequence_number = sequence_number
		self.__key_modified['Sequence_Number'] = 1

	def get_id(self):
		"""
		The method to get the id

		Returns:
			int: An int representing the id
		"""

		return self.__id

	def set_id(self, id):
		"""
		The method to set the value to id

		Parameters:
			id (int) : An int representing the id
		"""

		if id is not None and not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		self.__id = id
		self.__key_modified['id'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Query(object):
	def __init__(self):
		"""Creates an instance of Query"""

		self.__module = None
		self.__cvid = None
		self.__fields = None
		self.__page = None
		self.__criteria = None
		self.__key_modified = dict()

	def get_module(self):
		"""
		The method to get the module

		Returns:
			Module: An instance of Module
		"""

		return self.__module

	def set_module(self, module):
		"""
		The method to set the value to module

		Parameters:
			module (Module) : An instance of Module
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.modules import Module
		except Exception:
			from ..modules import Module

		if module is not None and not isinstance(module, Module):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: module EXPECTED TYPE: Module', None, None)
		
		self.__module = module
		self.__key_modified['module'] = 1

	def get_cvid(self):
		"""
		The method to get the cvid

		Returns:
			string: A string representing the cvid
		"""

		return self.__cvid

	def set_cvid(self, cvid):
		"""
		The method to set the value to cvid

		Parameters:
			cvid (string) : A string representing the cvid
		"""

		if cvid is not None and not isinstance(cvid, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: cvid EXPECTED TYPE: str', None, None)
		
		self.__cvid = cvid
		self.__key_modified['cvid'] = 1

	def get_fields(self):
		"""
		The method to get the fields

		Returns:
			list: An instance of list
		"""

		return self.__fields

	def set_fields(self, fields):
		"""
		The method to set the value to fields

		Parameters:
			fields (list) : An instance of list
		"""

		if fields is not None and not isinstance(fields, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: fields EXPECTED TYPE: list', None, None)
		
		self.__fields = fields
		self.__key_modified['fields'] = 1

	def get_page(self):
		"""
		The method to get the page

		Returns:
			int: An int representing the page
		"""

		return self.__page

	def set_page(self, page):
		"""
		The method to set the value to page

		Parameters:
			page (int) : An int representing the page
		"""

		if page is not None and not isinstance(page, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: page EXPECTED TYPE: int', None, None)
		
		self.__page = page
		self.__key_modified['page'] = 1

	def get_criteria(self):
		"""
		The method to get the criteria

		Returns:
			Criteria: An instance of Criteria
		"""

		return self.__criteria

	def set_criteria(self, criteria):
		"""
		The method to set the value to criteria

		Parameters:
			criteria (Criteria) : An instance of Criteria
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.bulk_read.criteria import Criteria
		except Exception:
			from .criteria import Criteria

		if criteria is not None and not isinstance(criteria, Criteria):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: criteria EXPECTED TYPE: Criteria', None, None)
		
		self.__criteria = criteria
		self.__key_modified['criteria'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



from .api_exception import APIException
from .response_handler import ResponseHandler
from .role import Role
from .response_wrapper import ResponseWrapper
from .roles_operations import RolesOperations



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Transition(object):
	def __init__(self):
		"""Creates an instance of Transition"""

		self.__next_transitions = None
		self.__parent_transition = None
		self.__percent_partial_save = None
		self.__data = None
		self.__next_field_value = None
		self.__name = None
		self.__criteria_matched = None
		self.__id = None
		self.__fields = None
		self.__criteria_message = None
		self.__type = None
		self.__execution_time = None
		self.__key_modified = dict()

	def get_next_transitions(self):
		"""
		The method to get the next_transitions

		Returns:
			list: An instance of list
		"""

		return self.__next_transitions

	def set_next_transitions(self, next_transitions):
		"""
		The method to set the value to next_transitions

		Parameters:
			next_transitions (list) : An instance of list
		"""

		if next_transitions is not None and not isinstance(next_transitions, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: next_transitions EXPECTED TYPE: list', None, None)
		
		self.__next_transitions = next_transitions
		self.__key_modified['next_transitions'] = 1

	def get_parent_transition(self):
		"""
		The method to get the parent_transition

		Returns:
			Transition: An instance of Transition
		"""

		return self.__parent_transition

	def set_parent_transition(self, parent_transition):
		"""
		The method to set the value to parent_transition

		Parameters:
			parent_transition (Transition) : An instance of Transition
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.blue_print.transition import Transition
		except Exception:
			from .transition import Transition

		if parent_transition is not None and not isinstance(parent_transition, Transition):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: parent_transition EXPECTED TYPE: Transition', None, None)
		
		self.__parent_transition = parent_transition
		self.__key_modified['parent_transition'] = 1

	def get_percent_partial_save(self):
		"""
		The method to get the percent_partial_save

		Returns:
			float: A float representing the percent_partial_save
		"""

		return self.__percent_partial_save

	def set_percent_partial_save(self, percent_partial_save):
		"""
		The method to set the value to percent_partial_save

		Parameters:
			percent_partial_save (float) : A float representing the percent_partial_save
		"""

		if percent_partial_save is not None and not isinstance(percent_partial_save, float):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: percent_partial_save EXPECTED TYPE: float', None, None)
		
		self.__percent_partial_save = percent_partial_save
		self.__key_modified['percent_partial_save'] = 1

	def get_data(self):
		"""
		The method to get the data

		Returns:
			Record: An instance of Record
		"""

		return self.__data

	def set_data(self, data):
		"""
		The method to set the value to data

		Parameters:
			data (Record) : An instance of Record
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.record import Record
		except Exception:
			from ..record import Record

		if data is not None and not isinstance(data, Record):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: data EXPECTED TYPE: Record', None, None)
		
		self.__data = data
		self.__key_modified['data'] = 1

	def get_next_field_value(self):
		"""
		The method to get the next_field_value

		Returns:
			string: A string representing the next_field_value
		"""

		return self.__next_field_value

	def set_next_field_value(self, next_field_value):
		"""
		The method to set the value to next_field_value

		Parameters:
			next_field_value (string) : A string representing the next_field_value
		"""

		if next_field_value is not None and not isinstance(next_field_value, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: next_field_value EXPECTED TYPE: str', None, None)
		
		self.__next_field_value = next_field_value
		self.__key_modified['next_field_value'] = 1

	def get_name(self):
		"""
		The method to get the name

		Returns:
			string: A string representing the name
		"""

		return self.__name

	def set_name(self, name):
		"""
		The method to set the value to name

		Parameters:
			name (string) : A string representing the name
		"""

		if name is not None and not isinstance(name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: name EXPECTED TYPE: str', None, None)
		
		self.__name = name
		self.__key_modified['name'] = 1

	def get_criteria_matched(self):
		"""
		The method to get the criteria_matched

		Returns:
			bool: A bool representing the criteria_matched
		"""

		return self.__criteria_matched

	def set_criteria_matched(self, criteria_matched):
		"""
		The method to set the value to criteria_matched

		Parameters:
			criteria_matched (bool) : A bool representing the criteria_matched
		"""

		if criteria_matched is not None and not isinstance(criteria_matched, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: criteria_matched EXPECTED TYPE: bool', None, None)
		
		self.__criteria_matched = criteria_matched
		self.__key_modified['criteria_matched'] = 1

	def get_id(self):
		"""
		The method to get the id

		Returns:
			int: An int representing the id
		"""

		return self.__id

	def set_id(self, id):
		"""
		The method to set the value to id

		Parameters:
			id (int) : An int representing the id
		"""

		if id is not None and not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		self.__id = id
		self.__key_modified['id'] = 1

	def get_fields(self):
		"""
		The method to get the fields

		Returns:
			list: An instance of list
		"""

		return self.__fields

	def set_fields(self, fields):
		"""
		The method to set the value to fields

		Parameters:
			fields (list) : An instance of list
		"""

		if fields is not None and not isinstance(fields, list):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: fields EXPECTED TYPE: list', None, None)
		
		self.__fields = fields
		self.__key_modified['fields'] = 1

	def get_criteria_message(self):
		"""
		The method to get the criteria_message

		Returns:
			string: A string representing the criteria_message
		"""

		return self.__criteria_message

	def set_criteria_message(self, criteria_message):
		"""
		The method to set the value to criteria_message

		Parameters:
			criteria_message (string) : A string representing the criteria_message
		"""

		if criteria_message is not None and not isinstance(criteria_message, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: criteria_message EXPECTED TYPE: str', None, None)
		
		self.__criteria_message = criteria_message
		self.__key_modified['criteria_message'] = 1

	def get_type(self):
		"""
		The method to get the type

		Returns:
			string: A string representing the type
		"""

		return self.__type

	def set_type(self, type):
		"""
		The method to set the value to type

		Parameters:
			type (string) : A string representing the type
		"""

		if type is not None and not isinstance(type, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: type EXPECTED TYPE: str', None, None)
		
		self.__type = type
		self.__key_modified['type'] = 1

	def get_execution_time(self):
		"""
		The method to get the execution_time

		Returns:
			datetime: An instance of datetime
		"""

		return self.__execution_time

	def set_execution_time(self, execution_time):
		"""
		The method to set the value to execution_time

		Parameters:
			execution_time (datetime) : An instance of datetime
		"""

		from datetime import datetime

		if execution_time is not None and not isinstance(execution_time, datetime):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: execution_time EXPECTED TYPE: datetime', None, None)
		
		self.__execution_time = execution_time
		self.__key_modified['execution_time'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class RelatedList(object):
	def __init__(self):
		"""Creates an instance of RelatedList"""

		self.__id = None
		self.__sequence_number = None
		self.__display_label = None
		self.__api_name = None
		self.__module = None
		self.__name = None
		self.__action = None
		self.__href = None
		self.__type = None
		self.__connectedmodule = None
		self.__linkingmodule = None
		self.__key_modified = dict()

	def get_id(self):
		"""
		The method to get the id

		Returns:
			int: An int representing the id
		"""

		return self.__id

	def set_id(self, id):
		"""
		The method to set the value to id

		Parameters:
			id (int) : An int representing the id
		"""

		if id is not None and not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		self.__id = id
		self.__key_modified['id'] = 1

	def get_sequence_number(self):
		"""
		The method to get the sequence_number

		Returns:
			string: A string representing the sequence_number
		"""

		return self.__sequence_number

	def set_sequence_number(self, sequence_number):
		"""
		The method to set the value to sequence_number

		Parameters:
			sequence_number (string) : A string representing the sequence_number
		"""

		if sequence_number is not None and not isinstance(sequence_number, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: sequence_number EXPECTED TYPE: str', None, None)
		
		self.__sequence_number = sequence_number
		self.__key_modified['sequence_number'] = 1

	def get_display_label(self):
		"""
		The method to get the display_label

		Returns:
			string: A string representing the display_label
		"""

		return self.__display_label

	def set_display_label(self, display_label):
		"""
		The method to set the value to display_label

		Parameters:
			display_label (string) : A string representing the display_label
		"""

		if display_label is not None and not isinstance(display_label, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: display_label EXPECTED TYPE: str', None, None)
		
		self.__display_label = display_label
		self.__key_modified['display_label'] = 1

	def get_api_name(self):
		"""
		The method to get the api_name

		Returns:
			string: A string representing the api_name
		"""

		return self.__api_name

	def set_api_name(self, api_name):
		"""
		The method to set the value to api_name

		Parameters:
			api_name (string) : A string representing the api_name
		"""

		if api_name is not None and not isinstance(api_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: api_name EXPECTED TYPE: str', None, None)
		
		self.__api_name = api_name
		self.__key_modified['api_name'] = 1

	def get_module(self):
		"""
		The method to get the module

		Returns:
			string: A string representing the module
		"""

		return self.__module

	def set_module(self, module):
		"""
		The method to set the value to module

		Parameters:
			module (string) : A string representing the module
		"""

		if module is not None and not isinstance(module, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: module EXPECTED TYPE: str', None, None)
		
		self.__module = module
		self.__key_modified['module'] = 1

	def get_name(self):
		"""
		The method to get the name

		Returns:
			string: A string representing the name
		"""

		return self.__name

	def set_name(self, name):
		"""
		The method to set the value to name

		Parameters:
			name (string) : A string representing the name
		"""

		if name is not None and not isinstance(name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: name EXPECTED TYPE: str', None, None)
		
		self.__name = name
		self.__key_modified['name'] = 1

	def get_action(self):
		"""
		The method to get the action

		Returns:
			string: A string representing the action
		"""

		return self.__action

	def set_action(self, action):
		"""
		The method to set the value to action

		Parameters:
			action (string) : A string representing the action
		"""

		if action is not None and not isinstance(action, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: action EXPECTED TYPE: str', None, None)
		
		self.__action = action
		self.__key_modified['action'] = 1

	def get_href(self):
		"""
		The method to get the href

		Returns:
			string: A string representing the href
		"""

		return self.__href

	def set_href(self, href):
		"""
		The method to set the value to href

		Parameters:
			href (string) : A string representing the href
		"""

		if href is not None and not isinstance(href, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: href EXPECTED TYPE: str', None, None)
		
		self.__href = href
		self.__key_modified['href'] = 1

	def get_type(self):
		"""
		The method to get the type

		Returns:
			string: A string representing the type
		"""

		return self.__type

	def set_type(self, type):
		"""
		The method to set the value to type

		Parameters:
			type (string) : A string representing the type
		"""

		if type is not None and not isinstance(type, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: type EXPECTED TYPE: str', None, None)
		
		self.__type = type
		self.__key_modified['type'] = 1

	def get_connectedmodule(self):
		"""
		The method to get the connectedmodule

		Returns:
			string: A string representing the connectedmodule
		"""

		return self.__connectedmodule

	def set_connectedmodule(self, connectedmodule):
		"""
		The method to set the value to connectedmodule

		Parameters:
			connectedmodule (string) : A string representing the connectedmodule
		"""

		if connectedmodule is not None and not isinstance(connectedmodule, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: connectedmodule EXPECTED TYPE: str', None, None)
		
		self.__connectedmodule = connectedmodule
		self.__key_modified['connectedmodule'] = 1

	def get_linkingmodule(self):
		"""
		The method to get the linkingmodule

		Returns:
			string: A string representing the linkingmodule
		"""

		return self.__linkingmodule

	def set_linkingmodule(self, linkingmodule):
		"""
		The method to set the value to linkingmodule

		Parameters:
			linkingmodule (string) : A string representing the linkingmodule
		"""

		if linkingmodule is not None and not isinstance(linkingmodule, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: linkingmodule EXPECTED TYPE: str', None, None)
		
		self.__linkingmodule = linkingmodule
		self.__key_modified['linkingmodule'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Org(object):
	def __init__(self):
		"""Creates an instance of Org"""

		self.__country = None
		self.__photo_id = None
		self.__city = None
		self.__description = None
		self.__mc_status = None
		self.__gapps_enabled = None
		self.__domain_name = None
		self.__translation_enabled = None
		self.__street = None
		self.__alias = None
		self.__currency = None
		self.__id = None
		self.__state = None
		self.__fax = None
		self.__employee_count = None
		self.__zip = None
		self.__website = None
		self.__currency_symbol = None
		self.__mobile = None
		self.__currency_locale = None
		self.__primary_zuid = None
		self.__zia_portal_id = None
		self.__time_zone = None
		self.__zgid = None
		self.__country_code = None
		self.__license_details = None
		self.__phone = None
		self.__company_name = None
		self.__privacy_settings = None
		self.__primary_email = None
		self.__iso_code = None
		self.__key_modified = dict()

	def get_country(self):
		"""
		The method to get the country

		Returns:
			string: A string representing the country
		"""

		return self.__country

	def set_country(self, country):
		"""
		The method to set the value to country

		Parameters:
			country (string) : A string representing the country
		"""

		if country is not None and not isinstance(country, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: country EXPECTED TYPE: str', None, None)
		
		self.__country = country
		self.__key_modified['country'] = 1

	def get_photo_id(self):
		"""
		The method to get the photo_id

		Returns:
			string: A string representing the photo_id
		"""

		return self.__photo_id

	def set_photo_id(self, photo_id):
		"""
		The method to set the value to photo_id

		Parameters:
			photo_id (string) : A string representing the photo_id
		"""

		if photo_id is not None and not isinstance(photo_id, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: photo_id EXPECTED TYPE: str', None, None)
		
		self.__photo_id = photo_id
		self.__key_modified['photo_id'] = 1

	def get_city(self):
		"""
		The method to get the city

		Returns:
			string: A string representing the city
		"""

		return self.__city

	def set_city(self, city):
		"""
		The method to set the value to city

		Parameters:
			city (string) : A string representing the city
		"""

		if city is not None and not isinstance(city, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: city EXPECTED TYPE: str', None, None)
		
		self.__city = city
		self.__key_modified['city'] = 1

	def get_description(self):
		"""
		The method to get the description

		Returns:
			string: A string representing the description
		"""

		return self.__description

	def set_description(self, description):
		"""
		The method to set the value to description

		Parameters:
			description (string) : A string representing the description
		"""

		if description is not None and not isinstance(description, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: description EXPECTED TYPE: str', None, None)
		
		self.__description = description
		self.__key_modified['description'] = 1

	def get_mc_status(self):
		"""
		The method to get the mc_status

		Returns:
			bool: A bool representing the mc_status
		"""

		return self.__mc_status

	def set_mc_status(self, mc_status):
		"""
		The method to set the value to mc_status

		Parameters:
			mc_status (bool) : A bool representing the mc_status
		"""

		if mc_status is not None and not isinstance(mc_status, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: mc_status EXPECTED TYPE: bool', None, None)
		
		self.__mc_status = mc_status
		self.__key_modified['mc_status'] = 1

	def get_gapps_enabled(self):
		"""
		The method to get the gapps_enabled

		Returns:
			bool: A bool representing the gapps_enabled
		"""

		return self.__gapps_enabled

	def set_gapps_enabled(self, gapps_enabled):
		"""
		The method to set the value to gapps_enabled

		Parameters:
			gapps_enabled (bool) : A bool representing the gapps_enabled
		"""

		if gapps_enabled is not None and not isinstance(gapps_enabled, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: gapps_enabled EXPECTED TYPE: bool', None, None)
		
		self.__gapps_enabled = gapps_enabled
		self.__key_modified['gapps_enabled'] = 1

	def get_domain_name(self):
		"""
		The method to get the domain_name

		Returns:
			string: A string representing the domain_name
		"""

		return self.__domain_name

	def set_domain_name(self, domain_name):
		"""
		The method to set the value to domain_name

		Parameters:
			domain_name (string) : A string representing the domain_name
		"""

		if domain_name is not None and not isinstance(domain_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: domain_name EXPECTED TYPE: str', None, None)
		
		self.__domain_name = domain_name
		self.__key_modified['domain_name'] = 1

	def get_translation_enabled(self):
		"""
		The method to get the translation_enabled

		Returns:
			bool: A bool representing the translation_enabled
		"""

		return self.__translation_enabled

	def set_translation_enabled(self, translation_enabled):
		"""
		The method to set the value to translation_enabled

		Parameters:
			translation_enabled (bool) : A bool representing the translation_enabled
		"""

		if translation_enabled is not None and not isinstance(translation_enabled, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: translation_enabled EXPECTED TYPE: bool', None, None)
		
		self.__translation_enabled = translation_enabled
		self.__key_modified['translation_enabled'] = 1

	def get_street(self):
		"""
		The method to get the street

		Returns:
			string: A string representing the street
		"""

		return self.__street

	def set_street(self, street):
		"""
		The method to set the value to street

		Parameters:
			street (string) : A string representing the street
		"""

		if street is not None and not isinstance(street, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: street EXPECTED TYPE: str', None, None)
		
		self.__street = street
		self.__key_modified['street'] = 1

	def get_alias(self):
		"""
		The method to get the alias

		Returns:
			string: A string representing the alias
		"""

		return self.__alias

	def set_alias(self, alias):
		"""
		The method to set the value to alias

		Parameters:
			alias (string) : A string representing the alias
		"""

		if alias is not None and not isinstance(alias, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: alias EXPECTED TYPE: str', None, None)
		
		self.__alias = alias
		self.__key_modified['alias'] = 1

	def get_currency(self):
		"""
		The method to get the currency

		Returns:
			string: A string representing the currency
		"""

		return self.__currency

	def set_currency(self, currency):
		"""
		The method to set the value to currency

		Parameters:
			currency (string) : A string representing the currency
		"""

		if currency is not None and not isinstance(currency, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: currency EXPECTED TYPE: str', None, None)
		
		self.__currency = currency
		self.__key_modified['currency'] = 1

	def get_id(self):
		"""
		The method to get the id

		Returns:
			int: An int representing the id
		"""

		return self.__id

	def set_id(self, id):
		"""
		The method to set the value to id

		Parameters:
			id (int) : An int representing the id
		"""

		if id is not None and not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		self.__id = id
		self.__key_modified['id'] = 1

	def get_state(self):
		"""
		The method to get the state

		Returns:
			string: A string representing the state
		"""

		return self.__state

	def set_state(self, state):
		"""
		The method to set the value to state

		Parameters:
			state (string) : A string representing the state
		"""

		if state is not None and not isinstance(state, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: state EXPECTED TYPE: str', None, None)
		
		self.__state = state
		self.__key_modified['state'] = 1

	def get_fax(self):
		"""
		The method to get the fax

		Returns:
			string: A string representing the fax
		"""

		return self.__fax

	def set_fax(self, fax):
		"""
		The method to set the value to fax

		Parameters:
			fax (string) : A string representing the fax
		"""

		if fax is not None and not isinstance(fax, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: fax EXPECTED TYPE: str', None, None)
		
		self.__fax = fax
		self.__key_modified['fax'] = 1

	def get_employee_count(self):
		"""
		The method to get the employee_count

		Returns:
			string: A string representing the employee_count
		"""

		return self.__employee_count

	def set_employee_count(self, employee_count):
		"""
		The method to set the value to employee_count

		Parameters:
			employee_count (string) : A string representing the employee_count
		"""

		if employee_count is not None and not isinstance(employee_count, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: employee_count EXPECTED TYPE: str', None, None)
		
		self.__employee_count = employee_count
		self.__key_modified['employee_count'] = 1

	def get_zip(self):
		"""
		The method to get the zip

		Returns:
			string: A string representing the zip
		"""

		return self.__zip

	def set_zip(self, zip):
		"""
		The method to set the value to zip

		Parameters:
			zip (string) : A string representing the zip
		"""

		if zip is not None and not isinstance(zip, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: zip EXPECTED TYPE: str', None, None)
		
		self.__zip = zip
		self.__key_modified['zip'] = 1

	def get_website(self):
		"""
		The method to get the website

		Returns:
			string: A string representing the website
		"""

		return self.__website

	def set_website(self, website):
		"""
		The method to set the value to website

		Parameters:
			website (string) : A string representing the website
		"""

		if website is not None and not isinstance(website, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: website EXPECTED TYPE: str', None, None)
		
		self.__website = website
		self.__key_modified['website'] = 1

	def get_currency_symbol(self):
		"""
		The method to get the currency_symbol

		Returns:
			string: A string representing the currency_symbol
		"""

		return self.__currency_symbol

	def set_currency_symbol(self, currency_symbol):
		"""
		The method to set the value to currency_symbol

		Parameters:
			currency_symbol (string) : A string representing the currency_symbol
		"""

		if currency_symbol is not None and not isinstance(currency_symbol, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: currency_symbol EXPECTED TYPE: str', None, None)
		
		self.__currency_symbol = currency_symbol
		self.__key_modified['currency_symbol'] = 1

	def get_mobile(self):
		"""
		The method to get the mobile

		Returns:
			string: A string representing the mobile
		"""

		return self.__mobile

	def set_mobile(self, mobile):
		"""
		The method to set the value to mobile

		Parameters:
			mobile (string) : A string representing the mobile
		"""

		if mobile is not None and not isinstance(mobile, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: mobile EXPECTED TYPE: str', None, None)
		
		self.__mobile = mobile
		self.__key_modified['mobile'] = 1

	def get_currency_locale(self):
		"""
		The method to get the currency_locale

		Returns:
			string: A string representing the currency_locale
		"""

		return self.__currency_locale

	def set_currency_locale(self, currency_locale):
		"""
		The method to set the value to currency_locale

		Parameters:
			currency_locale (string) : A string representing the currency_locale
		"""

		if currency_locale is not None and not isinstance(currency_locale, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: currency_locale EXPECTED TYPE: str', None, None)
		
		self.__currency_locale = currency_locale
		self.__key_modified['currency_locale'] = 1

	def get_primary_zuid(self):
		"""
		The method to get the primary_zuid

		Returns:
			string: A string representing the primary_zuid
		"""

		return self.__primary_zuid

	def set_primary_zuid(self, primary_zuid):
		"""
		The method to set the value to primary_zuid

		Parameters:
			primary_zuid (string) : A string representing the primary_zuid
		"""

		if primary_zuid is not None and not isinstance(primary_zuid, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: primary_zuid EXPECTED TYPE: str', None, None)
		
		self.__primary_zuid = primary_zuid
		self.__key_modified['primary_zuid'] = 1

	def get_zia_portal_id(self):
		"""
		The method to get the zia_portal_id

		Returns:
			string: A string representing the zia_portal_id
		"""

		return self.__zia_portal_id

	def set_zia_portal_id(self, zia_portal_id):
		"""
		The method to set the value to zia_portal_id

		Parameters:
			zia_portal_id (string) : A string representing the zia_portal_id
		"""

		if zia_portal_id is not None and not isinstance(zia_portal_id, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: zia_portal_id EXPECTED TYPE: str', None, None)
		
		self.__zia_portal_id = zia_portal_id
		self.__key_modified['zia_portal_id'] = 1

	def get_time_zone(self):
		"""
		The method to get the time_zone

		Returns:
			string: A string representing the time_zone
		"""

		return self.__time_zone

	def set_time_zone(self, time_zone):
		"""
		The method to set the value to time_zone

		Parameters:
			time_zone (string) : A string representing the time_zone
		"""

		if time_zone is not None and not isinstance(time_zone, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: time_zone EXPECTED TYPE: str', None, None)
		
		self.__time_zone = time_zone
		self.__key_modified['time_zone'] = 1

	def get_zgid(self):
		"""
		The method to get the zgid

		Returns:
			string: A string representing the zgid
		"""

		return self.__zgid

	def set_zgid(self, zgid):
		"""
		The method to set the value to zgid

		Parameters:
			zgid (string) : A string representing the zgid
		"""

		if zgid is not None and not isinstance(zgid, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: zgid EXPECTED TYPE: str', None, None)
		
		self.__zgid = zgid
		self.__key_modified['zgid'] = 1

	def get_country_code(self):
		"""
		The method to get the country_code

		Returns:
			string: A string representing the country_code
		"""

		return self.__country_code

	def set_country_code(self, country_code):
		"""
		The method to set the value to country_code

		Parameters:
			country_code (string) : A string representing the country_code
		"""

		if country_code is not None and not isinstance(country_code, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: country_code EXPECTED TYPE: str', None, None)
		
		self.__country_code = country_code
		self.__key_modified['country_code'] = 1

	def get_license_details(self):
		"""
		The method to get the license_details

		Returns:
			LicenseDetails: An instance of LicenseDetails
		"""

		return self.__license_details

	def set_license_details(self, license_details):
		"""
		The method to set the value to license_details

		Parameters:
			license_details (LicenseDetails) : An instance of LicenseDetails
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.org.license_details import LicenseDetails
		except Exception:
			from .license_details import LicenseDetails

		if license_details is not None and not isinstance(license_details, LicenseDetails):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: license_details EXPECTED TYPE: LicenseDetails', None, None)
		
		self.__license_details = license_details
		self.__key_modified['license_details'] = 1

	def get_phone(self):
		"""
		The method to get the phone

		Returns:
			string: A string representing the phone
		"""

		return self.__phone

	def set_phone(self, phone):
		"""
		The method to set the value to phone

		Parameters:
			phone (string) : A string representing the phone
		"""

		if phone is not None and not isinstance(phone, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: phone EXPECTED TYPE: str', None, None)
		
		self.__phone = phone
		self.__key_modified['phone'] = 1

	def get_company_name(self):
		"""
		The method to get the company_name

		Returns:
			string: A string representing the company_name
		"""

		return self.__company_name

	def set_company_name(self, company_name):
		"""
		The method to set the value to company_name

		Parameters:
			company_name (string) : A string representing the company_name
		"""

		if company_name is not None and not isinstance(company_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: company_name EXPECTED TYPE: str', None, None)
		
		self.__company_name = company_name
		self.__key_modified['company_name'] = 1

	def get_privacy_settings(self):
		"""
		The method to get the privacy_settings

		Returns:
			bool: A bool representing the privacy_settings
		"""

		return self.__privacy_settings

	def set_privacy_settings(self, privacy_settings):
		"""
		The method to set the value to privacy_settings

		Parameters:
			privacy_settings (bool) : A bool representing the privacy_settings
		"""

		if privacy_settings is not None and not isinstance(privacy_settings, bool):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: privacy_settings EXPECTED TYPE: bool', None, None)
		
		self.__privacy_settings = privacy_settings
		self.__key_modified['privacy_settings'] = 1

	def get_primary_email(self):
		"""
		The method to get the primary_email

		Returns:
			string: A string representing the primary_email
		"""

		return self.__primary_email

	def set_primary_email(self, primary_email):
		"""
		The method to set the value to primary_email

		Parameters:
			primary_email (string) : A string representing the primary_email
		"""

		if primary_email is not None and not isinstance(primary_email, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: primary_email EXPECTED TYPE: str', None, None)
		
		self.__primary_email = primary_email
		self.__key_modified['primary_email'] = 1

	def get_iso_code(self):
		"""
		The method to get the iso_code

		Returns:
			string: A string representing the iso_code
		"""

		return self.__iso_code

	def set_iso_code(self, iso_code):
		"""
		The method to set the value to iso_code

		Parameters:
			iso_code (string) : A string representing the iso_code
		"""

		if iso_code is not None and not isinstance(iso_code, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: iso_code EXPECTED TYPE: str', None, None)
		
		self.__iso_code = iso_code
		self.__key_modified['iso_code'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
    import logging
    import threading
    import datetime
    import time
    import os
    import json
    import zlib
    import base64
    import re
    from zcrmsdk.src.com.zoho.crm.api.util import Constants
    from zcrmsdk.src.com.zoho.crm.api.initializer import Initializer
    from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
    from zcrmsdk.src.com.zoho.crm.api.header_map import HeaderMap

except Exception:
    import logging
    import threading
    import os
    import json
    import time
    import datetime
    import zlib
    import base64
    import re
    from .constants import Constants
    from ..initializer import Initializer
    from ..exception import SDKException
    from ..header_map import HeaderMap


class Utility(object):
    """
    This class handles module field details.
    """

    apitype_vs_datatype = {}
    apitype_vs_structurename = {}
    new_file = False
    get_modified_modules = False
    force_refresh = False
    lock = threading.RLock()
    logger = logging.getLogger('SDKLogger')

    @staticmethod
    def set_handler_api_path(module_api_name, handler_instance):

        if handler_instance is None:
            return

        api_path = handler_instance.get_api_path()

        if module_api_name.lower() in api_path.lower():
            api_path_split = api_path.split("/")

            for i in range(len(api_path_split)):
                split_lower = api_path_split[i].lower()
                if split_lower == module_api_name.lower():
                    api_path_split[i] = module_api_name
                elif split_lower in Constants.DEFAULT_MODULENAME_VS_APINAME and Constants.DEFAULT_MODULENAME_VS_APINAME.get(split_lower) is not None:
                    api_path_split[i] = Constants.DEFAULT_MODULENAME_VS_APINAME.get(split_lower)

            api_path = "/".join(api_path_split)
            handler_instance.set_api_path(api_path)

    @staticmethod
    def get_fields(module_api_name, handler_instance=None):

        """
        This method to fetch field details of the current module for the current user and store the result in a JSON file.

        Parameters:
            module_api_name (str) : A string containing the CRM module API name.
            handler_instance (CommonAPIHandler) : A CommonAPIHandler instance.
        """

        try:
            from zcrmsdk.src.com.zoho.crm.api.initializer import Initializer
        except Exception:
            from ..initializer import Initializer

        last_modified_time = None

        try:
            with Utility.lock:
                resources_path = os.path.join(Initializer.get_initializer().resource_path,
                                              Constants.FIELD_DETAILS_DIRECTORY)

                if not os.path.exists(resources_path):
                    if module_api_name is not None and Utility.search_json_details(module_api_name) is not None:
                        return
                    os.makedirs(resources_path)

                module_api_name = Utility.verify_module_api_name(module_api_name)

                Utility.set_handler_api_path(module_api_name, handler_instance)

                record_field_details_path = Utility.get_file_name()

                if os.path.exists(record_field_details_path):
                    Utility.file_exists_flow(module_api_name, record_field_details_path, last_modified_time)
                elif Initializer.get_initializer().sdk_config.get_auto_refresh_fields():
                    Utility.new_file = True
                    Utility.fill_data_type()
                    api_supported_modules = Utility.get_modules(None)
                    record_field_details_json = Initializer.get_json(record_field_details_path) if os.path.exists(record_field_details_path) else {}

                    record_field_details_json[Constants.FIELDS_LAST_MODIFIED_TIME] = time.time() * 1000

                    for module in api_supported_modules:
                        if module.lower() not in record_field_details_json:
                            record_field_details_json[module.lower()] = {}
                            Utility.write_to_file(file_path=record_field_details_path,
                                                  file_contents=record_field_details_json)
                            field_details = Utility.get_fields_details(api_supported_modules[module][Constants.API_NAME])
                            record_field_details_json = Initializer.get_json(record_field_details_path)
                            record_field_details_json[module.lower()] = field_details
                            Utility.write_to_file(file_path=record_field_details_path,
                                                  file_contents=record_field_details_json)

                    Utility.new_file = False

                elif Utility.force_refresh and not Utility.get_modified_modules:
                    Utility.get_modified_modules = True
                    record_field_details_json = {}
                    Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)
                    Utility.modify_fields(record_field_details_path, last_modified_time)
                    Utility.get_modified_modules = False

                else:
                    Utility.fill_data_type()
                    record_field_details_json = {module_api_name.lower(): {}}
                    Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)
                    field_details = Utility.get_fields_details(module_api_name)
                    record_field_details_json = Initializer.get_json(record_field_details_path)
                    record_field_details_json[module_api_name.lower()] = field_details
                    Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)

        except Exception as e:
            if record_field_details_path is not None and os.path.exists(record_field_details_path):
                try:
                    record_field_details_json = Initializer.get_json(record_field_details_path)
                    if module_api_name is not None and module_api_name.lower() in record_field_details_json:
                        del record_field_details_json[module_api_name.lower()]

                    if Utility.new_file:
                        if Constants.FIELDS_LAST_MODIFIED_TIME in record_field_details_json:
                            del record_field_details_json[Constants.FIELDS_LAST_MODIFIED_TIME]
                        Utility.new_file = False

                    if Utility.get_modified_modules or Utility.force_refresh:
                        Utility.get_modified_modules = False
                        Utility.force_refresh = False

                        if last_modified_time is not None:
                            record_field_details_json[Constants.FIELDS_LAST_MODIFIED_TIME] = last_modified_time

                    Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)
                except Exception as ex:
                    sdk_exception = SDKException(cause=ex)
                    Utility.logger.error(Constants.EXCEPTION + sdk_exception.__str__())
                    raise sdk_exception

            if not isinstance(e, SDKException):
                e = SDKException(cause=e)
            Utility.logger.info(Constants.EXCEPTION + e.__str__())

            raise e

    @staticmethod
    def modify_fields(record_field_details_path, modified_time):
        modified_modules = Utility.get_modules(modified_time)
        record_field_details_json = Initializer.get_json(record_field_details_path)
        record_field_details_json[Constants.FIELDS_LAST_MODIFIED_TIME] = time.time() * 1000
        Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)
        if len(modified_modules) > 0:
            for module in modified_modules.keys():
                if module.lower() in record_field_details_json:
                    Utility.delete_fields(record_field_details_json, module)

            Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)

            for module in modified_modules.keys():
                module_data = modified_modules[module]
                Utility.get_fields(module_data[Constants.API_NAME])

    @staticmethod
    def file_exists_flow(module_api_name, record_field_details_path, last_modified_time):
        with Utility.lock:
            record_field_details_json = Initializer.get_json(record_field_details_path)

            if Initializer.get_initializer().sdk_config.get_auto_refresh_fields() and not Utility.new_file and not Utility.get_modified_modules and (Constants.FIELDS_LAST_MODIFIED_TIME not in record_field_details_json or Utility.force_refresh or (time.time() * 1000 - record_field_details_json[Constants.FIELDS_LAST_MODIFIED_TIME]) > 3600000):
                Utility.get_modified_modules = True
                last_modified_time = record_field_details_json[Constants.FIELDS_LAST_MODIFIED_TIME] if not Utility.force_refresh and Constants.FIELDS_LAST_MODIFIED_TIME in record_field_details_json else None
                Utility.modify_fields(record_field_details_path, last_modified_time)
                Utility.get_modified_modules = False

            elif not Initializer.get_initializer().sdk_config.get_auto_refresh_fields() and Utility.force_refresh and not Utility.get_modified_modules:
                Utility.get_modified_modules = True
                Utility.modify_fields(record_field_details_path, last_modified_time)
                Utility.get_modified_modules = False

            record_field_details_json = Initializer.get_json(record_field_details_path)

            if module_api_name is None or module_api_name.lower() in record_field_details_json:
                return

            else:
                Utility.fill_data_type()
                record_field_details_json[module_api_name.lower()] = {}
                Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)
                field_details = Utility.get_fields_details(module_api_name)
                record_field_details_json = Initializer.get_json(record_field_details_path)
                record_field_details_json[module_api_name.lower()] = field_details
                Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)

    @staticmethod
    def verify_module_api_name(module_api_name):
        if module_api_name is not None and Constants.DEFAULT_MODULENAME_VS_APINAME.get(module_api_name.lower()) is not None:
            return Constants.DEFAULT_MODULENAME_VS_APINAME.get(module_api_name.lower())

        record_field_details_path = Utility.get_file_name()
        if os.path.exists(record_field_details_path):
            fields_json = Initializer.get_json(record_field_details_path)
            if Constants.SDK_MODULE_METADATA in fields_json and module_api_name.lower() in fields_json[Constants.SDK_MODULE_METADATA]:
                return fields_json[Constants.SDK_MODULE_METADATA][module_api_name.lower()][Constants.API_NAME]

        return module_api_name

    @staticmethod
    def delete_fields(record_field_details_json, module):
        subform_modules = []
        fields_json = record_field_details_json[module.lower()]
        for key, value in fields_json.items():
            if Constants.SUBFORM in value and value[Constants.SUBFORM] and (value[Constants.MODULE]).lower() in record_field_details_json:
                subform_modules.append(value[Constants.MODULE])

        del record_field_details_json[module.lower()]

        if len(subform_modules) > 0:
            for subform_module in subform_modules:
                Utility.delete_fields(record_field_details_json, subform_module)

    @staticmethod
    def get_file_name():
        import zcrmsdk.src.com.zoho.crm.api.util.converter as Converter

        return os.path.join(Initializer.get_initializer().resource_path, Constants.FIELD_DETAILS_DIRECTORY,Converter.Converter.get_encoded_file_name())

    @staticmethod
    def get_related_lists(related_module_name, module_api_name, common_api_handler):
        with Utility.lock:
            try:
                is_new_data = False
                key = (module_api_name + Constants.UNDERSCORE + Constants.RELATED_LISTS).lower()
                resources_path = os.path.join(Initializer.get_initializer().resource_path,Constants.FIELD_DETAILS_DIRECTORY)

                if not os.path.exists(resources_path):
                    os.makedirs(resources_path)
                record_field_details_path = Utility.get_file_name()

                if not os.path.exists(record_field_details_path) or (
                        os.path.exists(record_field_details_path) and key not in Initializer.get_json(
                    record_field_details_path)):
                    is_new_data = True
                    related_list_values = Utility.get_related_list_details(module_api_name)
                    record_field_details_json = Initializer.get_json(record_field_details_path) if os.path.exists(record_field_details_path) else {}
                    record_field_details_json[key] = related_list_values
                    Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)

                record_field_details_json = Initializer.get_json(record_field_details_path)
                module_related_list = record_field_details_json[key]

                if not Utility.check_related_list_exists(related_module_name, module_related_list,common_api_handler) and not is_new_data:
                    del record_field_details_json[key]
                    Utility.write_to_file(file_path=record_field_details_path, file_contents=record_field_details_json)

                    Utility.get_related_lists(related_module_name, module_api_name, common_api_handler)

            except SDKException as e:
                Utility.logger.error(Constants.EXCEPTION + e.__str__())
                raise e

            except Exception as e:
                sdk_exception = SDKException(cause=e)
                Utility.logger.error(Constants.EXCEPTION + sdk_exception.__str__())
                raise sdk_exception

    @staticmethod
    def check_related_list_exists(related_module_name, module_related_list_array, common_api_handler):
        for related_list_jo in module_related_list_array:
            if related_list_jo[Constants.API_NAME] is not None and related_list_jo[Constants.API_NAME].lower() == related_module_name.lower():
                if related_list_jo[Constants.HREF] == Constants.NULL_VALUE:
                    raise SDKException(code=Constants.UNSUPPORTED_IN_API,
                                       message=common_api_handler.get_http_method() + ' ' + common_api_handler.get_api_path() + Constants.UNSUPPORTED_IN_API_MESSAGE)

                if related_list_jo[Constants.MODULE] != Constants.NULL_VALUE:
                    common_api_handler.set_module_api_name(related_list_jo[Constants.MODULE])
                    Utility.get_fields(related_list_jo[Constants.MODULE])
                return True

        return False

    @staticmethod
    def verify_photo_support(module_api_name):
        with Utility.lock:
            try:
                module_api_name = Utility.verify_module_api_name(module_api_name)
                if module_api_name in Constants.PHOTO_SUPPORTED_MODULES:
                    return True

                modules = Utility.get_module_names()
                if module_api_name.lower() in modules and modules[module_api_name.lower()] is not None:
                    module_meta_data = modules[module_api_name.lower()]

                    if module_meta_data[Constants.GENERATED_TYPE] != Constants.GENERATED_TYPE_CUSTOM:
                        raise SDKException(code=Constants.UPLOAD_PHOTO_UNSUPPORTED_ERROR,
                                           message=Constants.UPLOAD_PHOTO_UNSUPPORTED_MESSAGE + module_api_name)
            except SDKException as e:
                raise e

            except Exception as e:
                sdk_exception = SDKException(message=Constants.EXCEPTION, cause=e)
                raise sdk_exception

            return True

    @staticmethod
    def get_module_names():
        module_data = {}
        resources_path = os.path.join(Initializer.get_initializer().resource_path,
                                      Constants.FIELD_DETAILS_DIRECTORY)
        if not os.path.exists(resources_path):
            os.makedirs(resources_path)

        record_field_details_path = Utility.get_file_name()

        is_null = False
        if os.path.exists(record_field_details_path):
            json = Initializer.get_json(record_field_details_path)
            if Constants.SDK_MODULE_METADATA not in json:
                is_null = True
            elif json[Constants.SDK_MODULE_METADATA] is None:
                is_null = True

        if not os.path.exists(record_field_details_path) or is_null:
            module_data = Utility.get_modules(None)
            Utility.write_module_meta_data(record_field_details_path, module_data)
            return module_data

        record_field_details_json = Initializer.get_json(record_field_details_path)
        module_data = record_field_details_json[Constants.SDK_MODULE_METADATA]
        return module_data

    @staticmethod
    def write_module_meta_data(record_field_details_path, module_data):
        field_details_json = Initializer.get_json(record_field_details_path) if os.path.exists(
            record_field_details_path) else {}
        field_details_json[Constants.SDK_MODULE_METADATA] = module_data
        Utility.write_to_file(record_field_details_path, field_details_json)

    @staticmethod
    def get_related_list_details(module_api_name):
        import zcrmsdk.src.com.zoho.crm.api.related_lists as RelatedLists

        related_list_array = []

        response = RelatedLists.RelatedListsOperations(module_api_name).get_related_lists()

        if response is not None:
            if response.get_status_code() == Constants.NO_CONTENT_STATUS_CODE:
                return related_list_array

            data_object = response.get_object()

            if data_object is not None:

                if isinstance(data_object, RelatedLists.ResponseWrapper):
                    related_lists = data_object.get_related_lists()

                    for related_list in related_lists:
                        related_list_detail = {
                            Constants.API_NAME: related_list.get_api_name(),
                            Constants.MODULE: related_list.get_module() if related_list.get_module() is not None else Constants.NULL_VALUE,
                            Constants.NAME: related_list.get_name(),
                            Constants.HREF: related_list.get_href() if related_list.get_href() is not None else Constants.NULL_VALUE
                        }
                        related_list_array.append(related_list_detail)

                elif isinstance(data_object, RelatedLists.APIException):
                    error_response = {
                        Constants.CODE: data_object.get_code().get_value(),
                        Constants.STATUS: data_object.get_status().get_value(),
                        Constants.MESSAGE: data_object.get_message().get_value()
                    }
                    raise SDKException(code=Constants.API_EXCEPTION, details=error_response)

                else:
                    error_response = {
                        Constants.CODE: response.get_status_code()
                    }
                    raise SDKException(code=Constants.API_EXCEPTION, details=error_response)

            else:
                error_response = {
                    Constants.CODE: response.get_status_code()
                }
                raise SDKException(code=Constants.API_EXCEPTION, details=error_response)

        return related_list_array

    @staticmethod
    def get_fields_details(module_api_name):

        """
        This method to get module field data from Zoho CRM.
        :param module_api_name: A str containing the CRM module API name.
        :return: A object representing the Zoho CRM module field details.
        """
        import zcrmsdk.src.com.zoho.crm.api.fields as Field

        fields_details = {}
        response = Field.FieldsOperations(module_api_name).get_fields()

        if response is not None:
            if response.get_status_code() == Constants.NO_CONTENT_STATUS_CODE:
                return fields_details

            response_object = response.get_object()

            if isinstance(response_object, Field.ResponseWrapper):
                fields = response_object.get_fields()

                for field in fields:
                    if field.get_api_name() in Constants.KEYS_TO_SKIP:
                        continue

                    field_detail = {}

                    Utility.set_data_type(field_detail, field, module_api_name)

                    fields_details[field.get_api_name()] = field_detail

                if module_api_name.lower() in Constants.INVENTORY_MODULES:
                    field_detail = {
                        Constants.NAME: Constants.LINE_TAX,
                        Constants.TYPE: Constants.LIST_NAMESPACE,
                        Constants.STRUCTURE_NAME: Constants.LINE_TAX_NAMESPACE
                    }
                    fields_details[Constants.LINE_TAX] = field_detail

                if module_api_name.lower() == Constants.NOTES:
                    field_detail = {
                        Constants.NAME: Constants.ATTACHMENTS,
                        Constants.TYPE: Constants.LIST_NAMESPACE,
                        Constants.STRUCTURE_NAME: Constants.ATTACHMENTS_NAMESPACE
                    }
                    fields_details[Constants.ATTACHMENTS] = field_detail

            elif isinstance(response_object, Field.APIException):
                error_response = {
                    Constants.CODE: response_object.get_code().get_value(),
                    Constants.STATUS: response_object.get_status().get_value(),
                    Constants.MESSAGE: response_object.get_message().get_value()
                }
                raise SDKException(code=Constants.API_EXCEPTION, details=error_response)

        else:
            error_response = {
                Constants.CODE: response.get_status_code()
            }
            raise SDKException(code=Constants.API_EXCEPTION, details=error_response)
        return fields_details

    @staticmethod
    def search_json_details(key):
        package_name = Constants.PACKAGE_NAMESPACE + 'record.' + key

        from zcrmsdk.src.com.zoho.crm.api.initializer import Initializer

        for class_key in Initializer.json_details.keys():
            if class_key == package_name:
                return_json = {
                    Constants.MODULEPACKAGENAME: class_key,
                    Constants.MODULEDETAILS: Initializer.json_details[class_key]
                }

                return return_json

        return None

    @staticmethod
    def get_modules(header):

        import zcrmsdk.src.com.zoho.crm.api.modules as Modules

        api_names = {}
        header_map = HeaderMap()
        if header is not None:
            header_value = datetime.datetime.fromtimestamp(header / 1000.0)
            header_map.add(Modules.GetModulesHeader.if_modified_since, header_value)

        response = Modules.ModulesOperations().get_modules(header_map)

        if response is not None:
            if response.get_status_code() in [Constants.NO_CONTENT_STATUS_CODE, Constants.NOT_MODIFIED_STATUS_CODE]:
                return api_names

            response_object = response.get_object()

            if isinstance(response_object, Modules.ResponseWrapper):
                modules = response_object.get_modules()
                for module in modules:
                    if module.get_api_supported():
                        module_details = dict()
                        module_details[Constants.API_NAME] = module.get_api_name()
                        module_details[Constants.GENERATED_TYPE] = module.get_generated_type().get_value()
                        api_names[module.get_api_name().lower()] = module_details

            elif isinstance(response_object, Modules.APIException):
                error_response = {
                    Constants.CODE: response_object.get_code().get_value(),
                    Constants.STATUS: response_object.get_status().get_value(),
                    Constants.MESSAGE: response_object.get_message().get_value()
                }
                raise SDKException(code=Constants.API_EXCEPTION, details=error_response)

        if Utility.force_refresh:
            try:
                Utility.write_module_meta_data(Utility.get_file_name(), api_names)
            except Exception as e:

                raise SDKException(message=Constants.EXCEPTION, cause=e)

        return api_names

    @staticmethod
    def refresh_modules():
        Utility.force_refresh = True
        Utility.get_fields(None)
        Utility.force_refresh = False

    @staticmethod
    def get_json_object(json, key):
        for key_in_json in json.keys():
            if key_in_json.lower() == key.lower():
                return json[key_in_json]

        return None

    @staticmethod
    def check_data_type(value, type):

        if value is None:
            return False
        if type.lower() == Constants.OBJECT.lower():
            return True
        type = Constants.DATA_TYPE.get(type)
        class_name = value.__class__
        if class_name == type:
            return True
        else:
            return False

    @staticmethod
    def set_data_type(field_detail, field, module_api_name):
        api_type = field.get_data_type()
        module = ''
        key_name = field.get_api_name()

        if field.get_system_mandatory() is not None and field.get_system_mandatory() and not (
                module_api_name.lower() == Constants.CALLS and key_name.lower() == Constants.CALL_DURATION):
            field_detail[Constants.REQUIRED] = True

        if key_name in Constants.KEY_VS_INVENTORY_MODULE and module_api_name.lower() == Constants.KEY_VS_INVENTORY_MODULE.get(
                key_name):
            field_detail[Constants.NAME] = key_name
            field_detail[Constants.TYPE] = Constants.LIST_NAMESPACE
            field_detail[Constants.STRUCTURE_NAME] = Constants.INVENTORY_LINE_ITEMS
            field_detail[Constants.SKIP_MANDATORY] = True

            return

        elif key_name.lower() == Constants.PRICING_DETAILS.lower() and module_api_name.lower() == Constants.PRICE_BOOKS:
            field_detail[Constants.NAME] = key_name
            field_detail[Constants.TYPE] = Constants.LIST_NAMESPACE
            field_detail[Constants.STRUCTURE_NAME] = Constants.PRICINGDETAILS
            field_detail[Constants.SKIP_MANDATORY] = True

            return

        elif key_name.lower() == Constants.PARTICIPANT_API_NAME.lower() and (
                module_api_name.lower() == Constants.EVENTS or module_api_name.lower() == Constants.ACTIVITIES):
            field_detail[Constants.NAME] = key_name
            field_detail[Constants.TYPE] = Constants.LIST_NAMESPACE
            field_detail[Constants.STRUCTURE_NAME] = Constants.PARTICIPANTS
            field_detail[Constants.SKIP_MANDATORY] = True

            return

        elif key_name.lower() == Constants.COMMENTS.lower() and (
                module_api_name.lower() == Constants.SOLUTIONS or module_api_name.lower() == Constants.CASES):
            field_detail[Constants.NAME] = key_name
            field_detail[Constants.TYPE] = Constants.LIST_NAMESPACE
            field_detail[Constants.STRUCTURE_NAME] = Constants.COMMENT_NAMESPACE
            field_detail[Constants.LOOKUP] = True

            return

        elif key_name.lower() == Constants.LAYOUT.lower():
            field_detail[Constants.NAME] = key_name
            field_detail[Constants.TYPE] = Constants.LAYOUT_NAMESPACE
            field_detail[Constants.STRUCTURE_NAME] = Constants.LAYOUT_NAMESPACE
            field_detail[Constants.LOOKUP] = True

            return

        elif (key_name.lower() == Constants.TERRITORIES.lower() or key_name.lower() == Constants.TERRITORY.lower()) and field.get_custom_field() is not None and not field.get_custom_field():
            field_detail[Constants.NAME] = key_name
            field_detail[Constants.TYPE] = Constants.LIST_NAMESPACE
            field_detail[Constants.STRUCTURE_NAME] = Constants.TERRITORY_NAMESPACE
            field_detail[Constants.LOOKUP] = True

            return
        elif key_name.lower() == Constants.TAX.lower() or module_api_name.lower() == Constants.PRODUCTS.lower() :
            field_detail[Constants.NAME] = key_name
            field_detail[Constants.TYPE] = Constants.LIST_NAMESPACE
            field_detail[Constants.STRUCTURE_NAME] = Constants.TAX_NAMESPACE

            return
        elif api_type in Utility.apitype_vs_datatype:
            field_detail[Constants.TYPE] = Utility.apitype_vs_datatype.get(api_type)

        elif api_type.lower() == Constants.FORMULA.lower():
            if field.get_formula() is not None:
                return_type = field.get_formula().get_return_type()
                if return_type in Utility.apitype_vs_datatype:
                    field_detail[Constants.TYPE] = Utility.apitype_vs_datatype.get(return_type)

            field_detail[Constants.READ_ONLY] = True

        else:
            return

        if Constants.LOOKUP in api_type.lower():
            field_detail[Constants.LOOKUP] = True

        if Constants.CONSENT_LOOKUP in api_type.lower():
            field_detail[Constants.SKIP_MANDATORY] = True

        if Constants.MULTI_SELECT_LOOKUP in api_type.lower():
            field_detail[Constants.SKIP_MANDATORY] = True
            if field.get_multiselectlookup() is not None:
                if field.get_multiselectlookup().get_linking_module() is not None:
                    linking_module = field.get_multiselectlookup().get_linking_module()
                    field_detail[Constants.MODULE] = linking_module
                    module = linking_module

            field_detail[Constants.SUBFORM] = True

        if Constants.MULTI_USER_LOOKUP in api_type.lower():
            field_detail[Constants.SKIP_MANDATORY] = True
            if field.get_multiuserlookup() is not None:
                if field.get_multiuserlookup().get_linking_module() is not None:
                    linking_module = field.get_multiuserlookup().get_linking_module()
                    field_detail[Constants.MODULE] = linking_module
                    module = linking_module

            field_detail[Constants.SUBFORM] = True

        if api_type in Utility.apitype_vs_structurename:
            field_detail[Constants.STRUCTURE_NAME] = Utility.apitype_vs_structurename.get(api_type)

        if api_type.lower() == Constants.PICKLIST and field.get_pick_list_values() is not None and len(
                field.get_pick_list_values()) > 0:
            field_detail[Constants.PICKLIST] = True
            values = list(map(lambda x: x.get_display_value(), field.get_pick_list_values()))
            field_detail[Constants.VALUES] = values

        if api_type == Constants.SUBFORM:
            module = field.get_subform().get_module()
            field_detail[Constants.MODULE] = module
            field_detail[Constants.SKIP_MANDATORY] = True
            field_detail[Constants.SUBFORM] = True

        if api_type == Constants.LOOKUP:
            module = field.get_lookup().get_module()

            if module is not None and not module == Constants.SE_MODULE:
                field_detail[Constants.MODULE] = module
                if module.lower() == Constants.ACCOUNTS and not field.get_custom_field():
                    field_detail[Constants.SKIP_MANDATORY] = True

            else:
                module = ''

            field_detail[Constants.LOOKUP] = True

        if len(module) > 0:
            Utility.get_fields(module)

        field_detail[Constants.NAME] = key_name

    @staticmethod
    def fill_data_type():
        if len(Utility.apitype_vs_datatype) > 0:
            return

        field_api_names_string = ["textarea", "text", "website", "email", "phone", "mediumtext",
                                  "profileimage", "autonumber"]

        field_api_names_integer = ['integer']

        field_api_names_boolean = ['boolean']

        field_api_names_double = ['double', 'percent', 'lookup', 'currency']

        field_api_names_long = ['long', 'bigint']

        field_api_names_field_file = ['fileupload']

        field_api_names_datetime = ['datetime', 'event_reminder']

        field_api_names_date = ['date']

        field_api_names_lookup = ['lookup']

        field_api_names_owner_lookup = ['ownerlookup', 'userlookup']

        field_api_names_multiuser_lookup = ['multiuserlookup']

        field_api_names_multimodule_lookup = ['multimodulelookup']

        field_api_names_picklist = ['picklist']

        field_api_names_multiselect_picklist = ['multiselectpicklist']

        field_api_names_subform = ['subform']

        field_api_name_task_remind_at = ['ALARM']

        field_api_name_recurring_activity = ['RRULE']

        field_api_name_reminder = ['multireminder']

        field_api_name_consent_lookup = ['consent_lookup']

        field_api_name_image_upload = ["imageupload"]

        field_api_name_multiselect_lookup = ["multiselectlookup"]

        for field_api_name in field_api_names_string:
            Utility.apitype_vs_datatype[field_api_name] = Constants.STRING_NAMESPACE

        for field_api_name in field_api_names_integer:
            Utility.apitype_vs_datatype[field_api_name] = Constants.INTEGER_NAMESPACE

        for field_api_name in field_api_names_boolean:
            Utility.apitype_vs_datatype[field_api_name] = Constants.BOOLEAN_NAMESPACE

        for field_api_name in field_api_names_double:
            Utility.apitype_vs_datatype[field_api_name] = Constants.DOUBLE_NAMESPACE

        for field_api_name in field_api_names_long:
            Utility.apitype_vs_datatype[field_api_name] = Constants.LONG_NAMESPACE

        for field_api_name in field_api_names_datetime:
            Utility.apitype_vs_datatype[field_api_name] = Constants.DATETIME_NAMESPACE

        for field_api_name in field_api_names_date:
            Utility.apitype_vs_datatype[field_api_name] = Constants.DATE_NAMESPACE

        for field_api_name in field_api_names_lookup:
            Utility.apitype_vs_datatype[field_api_name] = Constants.RECORD_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.RECORD_NAMESPACE

        for field_api_name in field_api_names_owner_lookup:
            Utility.apitype_vs_datatype[field_api_name] = Constants.USER_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.USER_NAMESPACE

        for field_api_name in field_api_names_multiuser_lookup:
            Utility.apitype_vs_datatype[field_api_name] = Constants.LIST_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.RECORD_NAMESPACE

        for field_api_name in field_api_names_multimodule_lookup:
            Utility.apitype_vs_datatype[field_api_name] = Constants.LIST_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.MODULE_NAMESPACE

        for field_api_name in field_api_names_picklist:
            Utility.apitype_vs_datatype[field_api_name] = Constants.CHOICE_NAMESPACE

        for field_api_name in field_api_names_multiselect_picklist:
            Utility.apitype_vs_datatype[field_api_name] = Constants.LIST_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.CHOICE_NAMESPACE

        for field_api_name in field_api_names_subform:
            Utility.apitype_vs_datatype[field_api_name] = Constants.LIST_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.RECORD_NAMESPACE

        for field_api_name in field_api_names_field_file:
            Utility.apitype_vs_datatype[field_api_name] = Constants.LIST_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.FIELD_FILE_NAMESPACE

        for field_api_name in field_api_name_task_remind_at:
            Utility.apitype_vs_datatype[field_api_name] = Constants.REMINDAT_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.REMINDAT_NAMESPACE

        for field_api_name in field_api_name_recurring_activity:
            Utility.apitype_vs_datatype[field_api_name] = Constants.RECURRING_ACTIVITY_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.RECURRING_ACTIVITY_NAMESPACE

        for field_api_name in field_api_name_reminder:
            Utility.apitype_vs_datatype[field_api_name] = Constants.LIST_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.REMINDER_NAMESPACE

        for field_api_name in field_api_name_consent_lookup:
            Utility.apitype_vs_datatype[field_api_name] = Constants.CONSENT_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.CONSENT_NAMESPACE

        for field_api_name in field_api_name_image_upload:
            Utility.apitype_vs_datatype[field_api_name] = Constants.LIST_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.IMAGEUPLOAD_NAMESPACE

        for field_api_name in field_api_name_multiselect_lookup:
            Utility.apitype_vs_datatype[field_api_name] = Constants.LIST_NAMESPACE
            Utility.apitype_vs_structurename[field_api_name] = Constants.RECORD_NAMESPACE

    @staticmethod
    def write_to_file(file_path, file_contents):
        with open(file_path, mode="w") as file:
            json.dump(file_contents, file)
            file.flush()
            file.close()



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class Territory(object):
	def __init__(self):
		"""Creates an instance of Territory"""

		self.__assigned = None
		self.__name = None
		self.__id = None
		self.__assigned_time = None
		self.__assigned_by = None
		self.__key_modified = dict()

	def get_assigned(self):
		"""
		The method to get the assigned

		Returns:
			string: A string representing the assigned
		"""

		return self.__assigned

	def set_assigned(self, assigned):
		"""
		The method to set the value to assigned

		Parameters:
			assigned (string) : A string representing the assigned
		"""

		if assigned is not None and not isinstance(assigned, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: assigned EXPECTED TYPE: str', None, None)
		
		self.__assigned = assigned
		self.__key_modified['$assigned'] = 1

	def get_name(self):
		"""
		The method to get the name

		Returns:
			string: A string representing the name
		"""

		return self.__name

	def set_name(self, name):
		"""
		The method to set the value to name

		Parameters:
			name (string) : A string representing the name
		"""

		if name is not None and not isinstance(name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: name EXPECTED TYPE: str', None, None)
		
		self.__name = name
		self.__key_modified['Name'] = 1

	def get_id(self):
		"""
		The method to get the id

		Returns:
			int: An int representing the id
		"""

		return self.__id

	def set_id(self, id):
		"""
		The method to set the value to id

		Parameters:
			id (int) : An int representing the id
		"""

		if id is not None and not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		self.__id = id
		self.__key_modified['id'] = 1

	def get_assigned_time(self):
		"""
		The method to get the assigned_time

		Returns:
			datetime: An instance of datetime
		"""

		return self.__assigned_time

	def set_assigned_time(self, assigned_time):
		"""
		The method to set the value to assigned_time

		Parameters:
			assigned_time (datetime) : An instance of datetime
		"""

		from datetime import datetime

		if assigned_time is not None and not isinstance(assigned_time, datetime):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: assigned_time EXPECTED TYPE: datetime', None, None)
		
		self.__assigned_time = assigned_time
		self.__key_modified['$assigned_time'] = 1

	def get_assigned_by(self):
		"""
		The method to get the assigned_by

		Returns:
			User: An instance of User
		"""

		return self.__assigned_by

	def set_assigned_by(self, assigned_by):
		"""
		The method to set the value to assigned_by

		Parameters:
			assigned_by (User) : An instance of User
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.users import User
		except Exception:
			from ..users import User

		if assigned_by is not None and not isinstance(assigned_by, User):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: assigned_by EXPECTED TYPE: User', None, None)
		
		self.__assigned_by = assigned_by
		self.__key_modified['$assigned_by'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
	from zcrmsdk.src.com.zoho.crm.api.exception import SDKException
	from zcrmsdk.src.com.zoho.crm.api.util import Constants
except Exception:
	from ..exception import SDKException
	from ..util import Constants


class DeletedRecord(object):
	def __init__(self):
		"""Creates an instance of DeletedRecord"""

		self.__deleted_by = None
		self.__id = None
		self.__display_name = None
		self.__type = None
		self.__created_by = None
		self.__deleted_time = None
		self.__key_modified = dict()

	def get_deleted_by(self):
		"""
		The method to get the deleted_by

		Returns:
			User: An instance of User
		"""

		return self.__deleted_by

	def set_deleted_by(self, deleted_by):
		"""
		The method to set the value to deleted_by

		Parameters:
			deleted_by (User) : An instance of User
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.users import User
		except Exception:
			from ..users import User

		if deleted_by is not None and not isinstance(deleted_by, User):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: deleted_by EXPECTED TYPE: User', None, None)
		
		self.__deleted_by = deleted_by
		self.__key_modified['deleted_by'] = 1

	def get_id(self):
		"""
		The method to get the id

		Returns:
			int: An int representing the id
		"""

		return self.__id

	def set_id(self, id):
		"""
		The method to set the value to id

		Parameters:
			id (int) : An int representing the id
		"""

		if id is not None and not isinstance(id, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: id EXPECTED TYPE: int', None, None)
		
		self.__id = id
		self.__key_modified['id'] = 1

	def get_display_name(self):
		"""
		The method to get the display_name

		Returns:
			string: A string representing the display_name
		"""

		return self.__display_name

	def set_display_name(self, display_name):
		"""
		The method to set the value to display_name

		Parameters:
			display_name (string) : A string representing the display_name
		"""

		if display_name is not None and not isinstance(display_name, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: display_name EXPECTED TYPE: str', None, None)
		
		self.__display_name = display_name
		self.__key_modified['display_name'] = 1

	def get_type(self):
		"""
		The method to get the type

		Returns:
			string: A string representing the type
		"""

		return self.__type

	def set_type(self, type):
		"""
		The method to set the value to type

		Parameters:
			type (string) : A string representing the type
		"""

		if type is not None and not isinstance(type, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: type EXPECTED TYPE: str', None, None)
		
		self.__type = type
		self.__key_modified['type'] = 1

	def get_created_by(self):
		"""
		The method to get the created_by

		Returns:
			User: An instance of User
		"""

		return self.__created_by

	def set_created_by(self, created_by):
		"""
		The method to set the value to created_by

		Parameters:
			created_by (User) : An instance of User
		"""

		try:
			from zcrmsdk.src.com.zoho.crm.api.users import User
		except Exception:
			from ..users import User

		if created_by is not None and not isinstance(created_by, User):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: created_by EXPECTED TYPE: User', None, None)
		
		self.__created_by = created_by
		self.__key_modified['created_by'] = 1

	def get_deleted_time(self):
		"""
		The method to get the deleted_time

		Returns:
			datetime: An instance of datetime
		"""

		return self.__deleted_time

	def set_deleted_time(self, deleted_time):
		"""
		The method to set the value to deleted_time

		Parameters:
			deleted_time (datetime) : An instance of datetime
		"""

		from datetime import datetime

		if deleted_time is not None and not isinstance(deleted_time, datetime):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: deleted_time EXPECTED TYPE: datetime', None, None)
		
		self.__deleted_time = deleted_time
		self.__key_modified['deleted_time'] = 1

	def is_key_modified(self, key):
		"""
		The method to check if the user has modified the given key

		Parameters:
			key (string) : A string representing the key

		Returns:
			int: An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if key in self.__key_modified:
			return self.__key_modified.get(key)
		
		return None

	def set_key_modified(self, key, modification):
		"""
		The method to mark the given key as modified

		Parameters:
			key (string) : A string representing the key
			modification (int) : An int representing the modification
		"""

		if key is not None and not isinstance(key, str):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: key EXPECTED TYPE: str', None, None)
		
		if modification is not None and not isinstance(modification, int):
			raise SDKException(Constants.DATA_TYPE_ERROR, 'KEY: modification EXPECTED TYPE: int', None, None)
		
		self.__key_modified[key] = modification



try:
    import re
    from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException
    from zcrmsdk.src.com.zoho.crm.api.util.constants import Constants

except Exception:
    import re
    from zcrmsdk.src.com.zoho.crm.api.exception.sdk_exception import SDKException
    from .util.constants import Constants


class UserSignature(object):

    """
    This class represents the Zoho CRM User.
    """

    def __init__(self, email):

        """
        Creates an UserSignature class instance with the specified user email.

        Parameters:
            email (str) : A string containing the CRM user email
        """

        error = {}

        if re.fullmatch(Constants.EMAIL_REGEX, email) is None:
            error[Constants.FIELD] = Constants.EMAIL
            error[Constants.EXPECTED_TYPE] = Constants.EMAIL

            raise SDKException(Constants.USER_SIGNATURE_ERROR, details=error)

        self.__email = email

    def get_email(self):
        """
        This is a getter method to get __email.

        Returns:
            string: A string representing __email
        """

        return self.__email
